from optimal_guide_finder import Azimuth_Model
from optimal_guide_finder.Azimuth_Model import predict as apd
import copy
import os
import numpy as np
from optimal_guide_finder.Azimuth_Model import util
import shutil
import pickle
import pylab as plt
import pandas
from optimal_guide_finder.Azimuth_Model import local_multiprocessing
from optimal_guide_finder.Azimuth_Model import load_data
from optimal_guide_finder.Azimuth_Model.features import featurization as feat

def check_feature_set_dims(feature_sets):
    F2 = None
    for set in list(feature_sets.keys()):
        F = feature_sets[set].shape[0]
        if F2 is None: F = F2
        assert F == F2, "not same # individuals for feature %s" % set

    assert feature_sets !={}, "features are empty, check learn_options"


def set_target(learn_options, classification):
    assert 'target_name' not in list(learn_options.keys()) or learn_options['target_name'] is not None, "changed it to be automatically set here"
    if not classification:
        learn_options["target_name"] = learn_options['rank-transformed target name']
        learn_options["training_metric"] = 'spearmanr'
        learn_options['ground_truth_label'] = learn_options['target_name']
    else:
        learn_options["target_name"] = learn_options['binary target name']
        learn_options["training_metric"] = 'AUC'
        learn_options['ground_truth_label'] = learn_options['binary target name']

    if learn_options["V"]==3:
        assert learn_options['target_name']=='score_drug_gene_rank' or learn_options['target_name']=='score_drug_gene_threshold', "cannot use raw scores when mergind data"
        assert learn_options["ground_truth_label"]=='score_drug_gene_rank' or learn_options["ground_truth_label"]=='score_drug_gene_threshold', "cannot use raw scores when mergind data"

    return learn_options

def GP_setup(learn_options, likelihood='gaussian', degree=3, set_target_fn=set_target):
    learn_options["method"] = "GPy"
    learn_options['kernel degree'] = degree

    if likelihood == 'warped':
        learn_options['warpedGP'] = True
    else:
        learn_options['warpedGP'] = False
    learn_options = set_target_fn(learn_options, classification=False)

    return learn_options

def SVC_setup(learn_options, likelihood='gaussian', degree=3,  set_target_fn=set_target):
    learn_options["method"] = "SVC"
    learn_options = set_target_fn(learn_options, classification=True)

    return learn_options

def L1_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=False)
    learn_options["method"] = "linreg"
    learn_options["penalty"] = "L1"
    learn_options["feature_select"] = False
    if "alpha" not in list(learn_options.keys()):
        learn_options["alpha"] = np.array([1e-6*pow(1.3,x) for x in range(0,100)])
    learn_options["loss"] = "squared"

    return learn_options

def L2_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=False)
    learn_options["method"] = "linreg"
    learn_options["penalty"] = "L2"
    learn_options["feature_select"] = False
    if "alpha" not in list(learn_options.keys()):
        learn_options["alpha"] = np.array([1e-6*pow(1.3,x) for x in range(0,100)])
    learn_options["loss"] = "squared"

    return learn_options

def mean_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=False)
    learn_options['method'] = 'mean'
    return learn_options

def random_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=False)
    learn_options['method'] = 'random'
    return learn_options

def elasticnet_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=False)
    learn_options["method"] = "linreg"
    learn_options["penalty"] = "EN"
    learn_options["feature_select"] = False
    learn_options["loss"] = "squared"
    if "alpha" not in list(learn_options.keys()):
        learn_options["alpha"] = np.array([1e-5*pow(2,x) for x in range(0,30)])
    return learn_options

def DNN_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=False)
    learn_options['method'] = 'DNN'
    learn_options['DNN target variable'] = 'score'#'score_drug_gene_quantized'
    # learn_options['DNN architecture'] = (119, 10, 10, 10, 2)
    return learn_options

def RF_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=False)
    learn_options['method'] = 'RandomForestRegressor'
    return learn_options

def doench_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=True)
    learn_options['method'] = 'doench'
    return learn_options

def sgrna_from_doench_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=False)
    learn_options['method'] = 'sgrna_from_doench'
    return learn_options

def linreg_setup(learn_options, set_target_fn=set_target):
    learn_options["method"] = "linreg"
    learn_options["penalty"] = "L1"
    learn_options["feature_select"] = False
    if "alpha" not in list(learn_options.keys()):
        learn_options["alpha"] = np.array([0.0])
    learn_options["loss"] = "squared"
    learn_options = set_target_fn(learn_options, classification=False)

    return learn_options

def logregL1_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=True)
    learn_options["method"] = "logregL1"
    learn_options["penalty"] = "L1"
    learn_options["feature_select"] = False
    if "alpha" not in list(learn_options.keys()):
        learn_options["alpha"] = np.array([1e-6*pow(1.3,x) for x in range(0,100)])
    if "fit_intercept" not in learn_options:
        learn_options["fit_intercept"] = True
    return learn_options

def LASSOs_ensemble_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=False)
    learn_options["method"] = "lasso_ensemble"
    learn_options["penalty"] = "L1"
    learn_options["feature_select"] = False
    if "alpha" not in list(learn_options.keys()):
        learn_options["alpha"] = np.array([1e-6*pow(1.3,x) for x in range(0,100)])
    learn_options["loss"] = "squared"

    return learn_options

def xu_et_al_setup(learn_options, set_target_fn=set_target):
    learn_options = set_target_fn(learn_options, classification=True)
    learn_options["method"] = "xu_et_al"

    return learn_options

def adaboost_setup(learn_options, num_estimators=100, max_depth=3, learning_rate=0.1, set_target_fn=set_target, model="AdaBoost"):
    """
    """
    learn_options = set_target_fn(learn_options, classification=False)
    if model=="AdaBoost":
        learn_options['method'] = "AdaBoostRegressor"
    elif model=="AdaBoostClassifier":
        learn_options['method'] = "AdaBoostClassifier"
    else:
        raise Exception("model must be either AdaBoost or AdaBoost Classifier")
    learn_options['adaboost_version'] = 'python' # "R" or "python"

    if 'adaboost_loss' not in list(learn_options.keys()) and model=="AdaBoostRegressor":
        learn_options['adaboost_loss'] = 'ls' # alternatives: "lad", "huber", "quantile", see scikit docs for details
    if 'adaboost_alpha' not in list(learn_options.keys()):
        learn_options['adaboost_alpha'] = 0.5 # this parameter is only used by the huber and quantile loss functions.

    if not learn_options['adaboost_CV']:
        learn_options['adaboost_learning_rate'] = learning_rate
        learn_options['adaboost_n_estimators'] = num_estimators
        learn_options['adaboost_max_depth'] = max_depth
    else:
        learn_options['adaboost_n_estimators'] = num_estimators

    return learn_options


def shared_setup(learn_options, order, test):
    if 'num_proc' not in list(learn_options.keys()):
        learn_options['num_proc'] = None
    if 'num_thread_per_proc' not in list(learn_options.keys()):
        learn_options['num_thread_per_proc'] = None

    num_proc = Azimuth_Model.local_multiprocessing.configure(TEST=test, num_proc=learn_options["num_proc"],
                                                num_thread_per_proc=learn_options["num_thread_per_proc"])
    learn_options["num_proc"] = num_proc

    learn_options["order"] = order  # gets used many places in code, not just here

    if "cv" not in list(learn_options.keys()):
        # if no CV preference is specified, use leave-one-gene-out
        learn_options["cv"] = "gene"

    if "normalize_features" not in list(learn_options.keys()):
        # if no CV preference is specified, use leave-one-gene-out
        learn_options["normalize_features"] = True

    if "weighted" not in list(learn_options.keys()):
        learn_options['weighted'] = None

    if "all pairs" not in list(learn_options.keys()):
        learn_options["all pairs"] = False

    if "include_known_pairs" not in list(learn_options.keys()):
        learn_options["include_known_pairs"] = False

    if "include_gene_guide_feature" not in list(learn_options.keys()):
        learn_options["include_gene_guide_feature"] = 0 #used as window size, so 0 is none

    #these should default to true to match experiments before they were options:
    if "gc_features" not in list(learn_options.keys()):
        learn_options["gc_features"] = True
    if "nuc_features" not in list(learn_options.keys()):
        learn_options["nuc_features"] = True

    if 'train_genes' not in list(learn_options.keys()):
        learn_options["train_genes"] = None
    if 'test_genes' not in list(learn_options.keys()):
        learn_options["test_genes"] = None

    if "num_proc" not in learn_options:
        learn_options["num_proc"] = None
    if "num_thread_per_proc" not in learn_options:
        learn_options["num_thread_per_proc"] = None

    if 'seed' not in learn_options:
        learn_options['seed'] = 1

    if "flipV1target" not in learn_options:
        learn_options["flipV1target"] = False

    if 'num_genes_remove_train' not in learn_options:
        learn_options['num_genes_remove_train'] = None

    if "include_microhomology" not in learn_options:
        learn_options["include_microhomology"] = False

    if "algorithm_hyperparam_search" not in learn_options:
        learn_options["algorithm_hyperparam_search"] = "grid" # other options is bo for bayesian optimization

    return num_proc

def setup(test=False, order=1, learn_options=None, data_file=None, pam_audit=True, length_audit=True):

    num_proc = shared_setup(learn_options, order, test)

    assert "testing_non_binary_target_name" in list(learn_options.keys()), "need this in order to get metrics, though used to be not needed, so you may newly see this error"
    if learn_options["testing_non_binary_target_name"] not in ['ranks', 'raw', 'thrs']:
        raise Exception('learn_otions["testing_non_binary_target_name"] must be in ["ranks", "raw", "thrs"]')

    Xdf, Y, gene_position, target_genes = Azimuth_Model.load_data.from_file(data_file, learn_options)
    learn_options['all_genes'] = target_genes

    if test:
        learn_options["order"] = 1

    if 'convert_30mer_to_31mer' in learn_options and learn_options['convert_30mer_to_31mer'] is True:
        print("WARNING!!! converting 30 mer to 31 mer (and then cutting off first nucleotide to go back to 30mer with a right shift)")
        for i in range(Xdf.shape[0]):
            Xdf['30mer'].iloc[i] = Azimuth_Model.util.convert_to_thirty_one(Xdf.iloc[i]["30mer"], Xdf.index.values[i][1], Xdf.iloc[i]["Strand"])
        # to_keep = Xdf['30mer'].isnull() == False
        # Xdf = Xdf[to_keep]
        # gene_position = gene_position[to_keep]
        # Y = Y[to_keep]
        Xdf["30mer"] = Xdf["30mer"].apply(lambda x: x[1:]) # chop the first nucleotide

    if 'left_right_guide_ind' in learn_options and learn_options['left_right_guide_ind'] is not None:
        seq_start, seq_end, expected_length = learn_options['left_right_guide_ind']
        assert len(Xdf["30mer"].values[0]) == expected_length
        Xdf['30mer'] = Xdf['30mer'].apply(lambda seq: seq[seq_start:seq_end])

    feature_sets = feat.featurize_data(Xdf, learn_options, Y, gene_position, pam_audit=pam_audit, length_audit=length_audit)
    np.random.seed(learn_options['seed'])

    return Y, feature_sets, target_genes, learn_options, num_proc


def run_models(models, orders, GP_likelihoods=['gaussian', 'warped'], WD_kernel_degrees=[3],
               adaboost_learning_rates=[0.1], adaboost_num_estimators=[100], adaboost_max_depths=[3],
               learn_options_set=None, test=False, CV=True, setup_function=setup, set_target_fn=set_target, pam_audit=True, length_audit=True, return_data=False):

    '''
    CV is set to false if want to train a final model and not cross-validate, but it goes in to what
    looks like cv code
    '''


    results = {}
    assert learn_options_set is not None, "need to specify learn_options_set"
    all_learn_options = {}

    #shorten so easier to display on graphs
    feat_models_short = {'L1':"L1", 'L2':"L2", 'elasticnet':"EN", 'linreg':"LR",
                         'RandomForest': "RF",
                         'AdaBoost':"AB", 'AdaBoostClassifier':"ABClass", 'doench': 'doench',
                         "logregL1": "logregL1", "sgrna_from_doench":"sgrna_from_doench", 'SVC': 'SVC', 'xu_et_al': 'xu_et_al'}

    if not CV:
        print("Received option CV=False, so I'm training using all of the data")
        assert len(list(learn_options_set.keys())) == 1, "when CV is False, only 1 set of learn options is allowed"
        assert len(models) == 1, "when CV is False, only 1 model is allowed"


    for learn_options_str in list(learn_options_set.keys()):
        # these options get augmented in setup
        partial_learn_opt = learn_options_set[learn_options_str]
        # if the model requires encoded features
        for model in models:
            # models requiring explicit featurization
            if model in list(feat_models_short.keys()):
                for order in orders:
                    print("running %s, order %d for %s" % (model, order, learn_options_str))

                    Y, feature_sets, target_genes, learn_options, num_proc = setup_function(test=test, order=order, learn_options=partial_learn_opt, pam_audit=pam_audit, length_audit=length_audit) # TODO precompute features for all orders, as this is repated for each model

                    if model == 'L1':
                        learn_options_model = L1_setup(copy.deepcopy(learn_options), set_target_fn=set_target_fn)
                    elif model == 'L2':
                        learn_options_model = L2_setup(copy.deepcopy(learn_options), set_target_fn=set_target_fn)
                    elif model == 'elasticnet':
                        learn_options_model = elasticnet_setup(copy.deepcopy(learn_options), set_target_fn=set_target_fn)
                    elif model == 'linreg':
                        learn_options_model = linreg_setup(copy.deepcopy(learn_options), set_target_fn=set_target_fn)
                    elif model == "logregL1":
                        learn_options_model = logregL1_setup(copy.deepcopy(learn_options), set_target_fn=set_target_fn)
                    elif model == 'RandomForest':
                        learn_options_model = RF_setup(copy.deepcopy(learn_options), set_target_fn=set_target_fn)
                    elif model == 'SVC':
                        learn_options_model = SVC_setup(copy.deepcopy(learn_options), set_target_fn=set_target_fn)
                    elif model == 'doench':
                        learn_options_model = doench_setup(copy.deepcopy(learn_options), set_target_fn=set_target_fn)
                    elif model == 'sgrna_from_doench':
                        learn_options_model = sgrna_from_doench_setup(copy.deepcopy(learn_options), set_target_fn=set_target_fn)
                    elif model == 'xu_et_al':
                        learn_options_model = xu_et_al_setup(copy.deepcopy(learn_options), set_target_fn=set_target_fn)
                    elif model == 'AdaBoost' or 'AdaBoostClassifier':
                        for learning_rate in adaboost_learning_rates:
                            for num_estimators in adaboost_num_estimators:
                                for max_depth in adaboost_max_depths:
                                    learn_options_model = adaboost_setup(copy.deepcopy(learn_options), learning_rate=learning_rate, num_estimators=num_estimators, max_depth=max_depth, set_target_fn=set_target_fn, model=model)
                        model_string = feat_models_short[model] + '_or%d_md%d_lr%.2f_n%d_%s' % (learn_options_set[learn_options_str]["order"], max_depth, learning_rate, num_estimators, learn_options_str)
                    if model != 'AdaBoost':
                        model_string = feat_models_short[model] + '_ord%d_%s' % (learn_options_set[learn_options_str]["order"], learn_options_str)

                    results[model_string] = apd.cross_validate(Y, feature_sets, learn_options=learn_options_model, TEST=test, CV=CV)

                    all_learn_options[model_string] = learn_options_model
            # if the model doesn't require explicit featurization
            else:
                assert setup_fn==setup, "not yet modified to handle this"
                print("running %s for %s" % (model, learn_options_str))
                Y, feature_sets, target_genes, learn_options, num_proc = setup(test=test, order=1, learn_options=partial_learn_opt, pam_audit=pam_audit, length_audit=length_audit)
                if model == 'mean':
                    learn_options_model = mean_setup(copy.deepcopy(learn_options))
                elif model == 'random':
                    learn_options_model = random_setup(copy.deepcopy(learn_options))
                elif model == 'DNN':
                    learn_options_model = DNN_setup(copy.deepcopy(learn_options))
                elif model == 'GP':
                    for likelihood in GP_likelihoods:
                        for degree in WD_kernel_degrees:
                            learn_options_model = GP_setup(copy.deepcopy(learn_options), likelihood=likelihood, degree=degree)
                            model_string = '%s_%s_degree%d_%s' % (model, likelihood, degree, learn_options_str)
                            results[model_string] = apd.cross_validate(Y, feature_sets, learn_options=learn_options_model,TEST=test, CV=CV)

                else:
                    raise NotImplementedError("model %s not supported" % model)

                # "GP" already calls apd.cross_validate() and has its own model_string, so skip this.
                if model != "GP":
                    model_string = model + '_%s' % learn_options_str
                    results[model_string] = apd.cross_validate(Y, feature_sets, learn_options=learn_options_model, TEST=test, CV=CV)

            all_learn_options[model_string] = learn_options_model

    return results, all_learn_options


def pickle_runner_results(exp_name, results, all_learn_options, relpath="/../" + "results"):
    abspath = os.path.abspath(__file__)
    dname = os.path.dirname(abspath) + relpath
    if not os.path.exists(dname):
        os.makedirs(dname)
        print("Created directory: %s" % str(dname))
    if exp_name is None:
        exp_name = list(results.keys())[0]
    myfile = dname+'/'+ exp_name + '.pickle'
    with open(myfile, 'wb') as f:
        print("writing results to %s" % myfile)
        pickle.dump((results, all_learn_options), f, -1)

def runner(models, learn_options, GP_likelihoods=None, orders=None, WD_kernel_degrees=None, where='local', cluster_user='fusi', cluster='RR1-N13-09-H44', test=False, exp_name = None, **kwargs):

    if where == 'local':
        results, all_learn_options = run_models(models, orders=orders, GP_likelihoods=GP_likelihoods, learn_options_set=learn_options, WD_kernel_degrees=WD_kernel_degrees, test=test, **kwargs)
        all_metrics, gene_names = Azimuth_Model.util.get_all_metrics(results, learn_options)
        Azimuth_Model.util.plot_all_metrics(all_metrics, gene_names, all_learn_options, save=True)

        # for non-local (i.e. cluster), the comparable code is in cli_run_model.py
        pickle_runner_results(exp_name, results, all_learn_options)

        return results, all_learn_options, all_metrics, gene_names

    elif where == 'cluster':
        from . import cluster_job

        # create random cluster directory, dump learn options, and create cluster file
        temapdir, user, clust_filename = cluster_job.create(cluster_user, models, orders, WD_kernel_degrees, GP_likelihoods, exp_name=exp_name, learn_options=learn_options, **kwargs)

        # raw_input("Submit job to HPC and press any key when it's finished: ")
        # util.plot_cluster_results(directory=temapdir)

        #stdout = temapdir + r"/stdout"
        #stderr = temapdir + r"/stderr"
        #if not os.path.exists(stdout): os.makedirs(stdout)
        #if not os.path.exists(stderr): os.makedirs(stderr)

        return temapdir, clust_filename, user#, stdout, stderr

def save_final_model_V3(filename=None, include_position=True, learn_options=None, short_name='final', pam_audit=True, length_audit=True):
    '''
    run_models(produce_final_model=True) is what saves the model
    '''
    test = False
    assert filename is not None, "need to provide filename to save final model"

    if learn_options is None:
        if include_position:
            learn_options = {"V": 3,
                        'train_genes': Azimuth_Model.load_data.get_V3_genes(),
                        'test_genes': Azimuth_Model.load_data.get_V3_genes(),
                        "testing_non_binary_target_name": 'ranks',
                        'include_pi_nuc_feat': True,
                        "gc_features": True,
                        "nuc_features": True,
                        "include_gene_position": True,
                        "include_NGGX_interaction": True,
                        "include_Tm": True,
                        "include_strand": False,
                        "include_gene_feature": False,
                        "include_gene_guide_feature": 0,
                        "extra pairs": False,
                        "weighted": None,
                        "training_metric": 'spearmanr',
                        "NDGC_k": 10,
                        "cv": "gene",
                        "include_gene_effect": False,
                        "include_drug": False,
                        "include_sgRNAscore": False,
                        'adaboost_loss' : 'ls', # main "ls", alternatives: "lad", "huber", "quantile", see scikit docs for details
                        'adaboost_alpha': 0.5, # this parameter is only used by the huber and quantile loss functions.
                        'normalize_features': False,
                        'adaboost_CV' : False
                        }
        else:
            learn_options = {"V": 3,
                'train_genes': Azimuth_Model.load_data.get_V3_genes(),
                'test_genes': Azimuth_Model.load_data.get_V3_genes(),
                "testing_non_binary_target_name": 'ranks',
                'include_pi_nuc_feat': True,
                "gc_features": True,
                "nuc_features": True,
                "include_gene_position": False,
                "include_NGGX_interaction": True,
                "include_Tm": True,
                "include_strand": False,
                "include_gene_feature": False,
                "include_gene_guide_feature": 0,
                "extra pairs": False,
                "weighted": None,
                "training_metric": 'spearmanr',
                "NDGC_k": 10,
                "cv": "gene",
                "include_gene_effect": False,
                "include_drug": False,
                "include_sgRNAscore": False,
                'adaboost_loss' : 'ls', # main "ls", alternatives: "lad", "huber", "quantile", see scikit docs for details
                'adaboost_alpha': 0.5, # this parameter is only used by the huber and quantile loss functions.
                'normalize_features': False,
                 'adaboost_CV' : False
                }

    learn_options_set = {short_name: learn_options}
    results, all_learn_options = run_models(["AdaBoost"], orders=[2], adaboost_learning_rates=[0.1],
                                            adaboost_max_depths=[3], adaboost_num_estimators=[100],
                                            learn_options_set=learn_options_set,
                                            test=test, CV=False, pam_audit=length_audit, length_audit=length_audit)
    model = list(results.values())[0][3][0]

    with open(filename, 'wb') as f:
        pickle.dump((model, learn_options), f, -1)

    return model


def predict(seq, aa_cut=None, percent_peptide=None, model=None, model_file=None, pam_audit=True, length_audit=False, learn_options_override=None):
    """
    Args:
        seq: numpy array of 30 nt sequences.
        aa_cut: numpy array of amino acid cut positions (optional).
        percent_peptide: numpy array of percent peptide (optional).
        model: model instance to use for prediction (optional).
        model_file: file name of pickled model to use for prediction (optional).
        pam_audit: check PAM of each sequence.
        length_audit: check length of each sequence.
        learn_options_override: a dictionary indicating which learn_options to override (optional).

    Returns: a numpy array of predictions.
    """
    # assert not (model is None and model_file is None), "you have to specify either a model or a model_file"
    assert isinstance(seq, (np.ndarray)), "Please ensure seq is a numpy array"
    assert len(seq[0]) > 0, "Make sure that seq is not empty"
    assert isinstance(seq[0], str), "Please ensure input sequences are in string format, i.e. 'AGAG' rather than ['A' 'G' 'A' 'G'] or alternate representations"

    if aa_cut is not None:
        assert len(aa_cut) > 0, "Make sure that aa_cut is not empty"
        assert isinstance(aa_cut, (np.ndarray)), "Please ensure aa_cut is a numpy array"
        assert np.all(np.isreal(aa_cut)), "amino-acid cut position needs to be a real number"

    if percent_peptide is not None:
        assert len(percent_peptide) > 0, "Make sure that percent_peptide is not empty"
        assert isinstance(percent_peptide, (np.ndarray)), "Please ensure percent_peptide is a numpy array"
        assert np.all(np.isreal(percent_peptide)), "percent_peptide needs to be a real number"


    if model_file is None:
        azimuth_saved_model_dir = os.path.join(os.path.dirname(Azimuth_Model.__file__), 'saved_models')
        # azimuth_saved_model_dir = os.path.join(os.path.dirname(Azimuth_Model.__file__), 'saved_models')
        if np.any(percent_peptide == -1) or (percent_peptide is None and aa_cut is None):
            #print("No model file specified, using V3_model_nopos")
            model_name = 'V3_model_nopos.pickle'
        else:
            print("No model file specified, using V3_model_full")
            model_name = 'V3_model_full.pickle'

        model_file = os.path.join(azimuth_saved_model_dir, model_name)

    if model is None:
        with open(model_file, 'rb') as f:
            model, learn_options = pickle.load(f, encoding='latin1')
    else:
        model, learn_options = model

    learn_options["V"] = 2

    learn_options = override_learn_options(learn_options_override, learn_options)

    # Y, feature_sets, target_genes, learn_options, num_proc = setup(test=False, order=2, learn_options=learn_options, data_file=test_filename)
    # inputs, dim, dimsum, feature_names = apd.concatenate_feature_sets(feature_sets)

    Xdf = pandas.DataFrame(columns=['30mer', 'Strand'], data=list(zip(seq, ['NA' for x in range(len(seq))])))

    if np.all(percent_peptide != -1) and (percent_peptide is not None and aa_cut is not None):
        gene_position = pandas.DataFrame(columns=['Percent Peptide', 'Amino Acid Cut position'], data=list(zip(percent_peptide, aa_cut)))
    else:
        gene_position = pandas.DataFrame(columns=['Percent Peptide', 'Amino Acid Cut position'], data=list(zip(np.ones(seq.shape[0])*-1, np.ones(seq.shape[0])*-1)))

    feature_sets = feat.featurize_data(Xdf, learn_options, pandas.DataFrame(), gene_position, pam_audit=pam_audit, length_audit=length_audit)
    inputs, dim, dimsum, feature_names = Azimuth_Model.util.concatenate_feature_sets(feature_sets)

    #print "CRISPR"
    #pandas.DataFrame(inputs).to_csv("CRISPR.inputs.test.csv")
    #import iapdb; iapdb.set_trace()

    # call to scikit-learn, returns a vector of predicted values
    preds = model.predict(inputs)

    # also check that predictions are not 0/1 from a classifier.predict() (instead of predict_proba() or decision_function())
    unique_preds = np.unique(preds)
    ok = False
    for pr in preds:
        if pr not in [0,1]:
            ok = True
    assert ok, "model returned only 0s and 1s"
    return preds

def override_learn_options(learn_options_override, learn_options):
    """
    override all keys seen in learn_options_override to alter learn_options
    """
    if learn_options_override is not None:
        for k in list(learn_options_override.keys()):
            learn_options[k] = learn_options_override[k]
    return learn_options

def fill_learn_options(learn_options_used_to_fill, learn_options_with_possible_missing):
    """
    only fill in keys that are missing from learn_options from learn_options_fill
    """
    if learn_options_used_to_fill is not None:
        for k in list(learn_options_used_to_fill.keys()):
            if k not in learn_options_with_possible_missing:
                learn_options_with_possible_missing[k] = learn_options_used_to_fill[k]
    return learn_options_with_possible_missing


def write_results(predictions, file_to_predict):
    newfile = file_to_predict.replace(".csv", ".pred.csv")
    data = pandas.read_csv(file_to_predict)
    data['predictions'] = predictions
    data.to_csv(newfile)
    print("wrote results to %s" % newfile)
    return data, newfile

if __name__ == '__main__':
    #save_final_model_V3(filename='Azimuth_Model/azure_models/V3_model_full.pickle', include_position=True)

    save_final_model_V3(filename='saved_models/V3_model_nopos.pickle', include_position=False)
    save_final_model_V3(filename='saved_models/V3_model_full.pickle', include_position=True)

    # predict('GGGCCGCTGTTGCAGGTGGCGGGTAGGATC', 'sense', 1200, 30.3, model_file='../saved_models/final_model_nicolo.pickle')


    learn_options = {"V": 3,
                "train_genes": Azimuth_Model.load_data.get_V3_genes(),
                "test_genes": Azimuth_Model.load_data.get_V3_genes(),
                "target_name": 'score_drug_gene_rank',
                "testing_non_binary_target_name": 'ranks',
                'include_pi_nuc_feat': True,
                "gc_features": True,
                "nuc_features": True,
                "include_gene_position": True,
                "include_NGGX_interaction": True,
                "include_Tm": True,
                "include_strand": False,
                "include_gene_feature": False,
                "include_gene_guide_feature": 0,
                "extra pairs": False,
                "weighted": None,
                "training_metric": 'spearmanr',
                "NDGC_k": 10,
                "cv": "gene",
                "adaboost_loss" : 'ls',
                "include_gene_effect": False,
                "include_drug": False,
                "include_sgRNAscore": False,
                'adaboost_loss' : 'ls', # main "ls", alternatives: "lad", "huber", "quantile", see scikit docs for details
                'adaboost_alpha': 0.5, # this parameter is only used by the huber and quantile loss functions.
                'adaboost_CV' : False
                }

    learn_options_set = {"post bug fix":learn_options}

    #runner(['AdaBoost'], learn_options_set, orders=[2], where='local', adaboost_learning_rates=[0.1],  adaboost_max_depths=[3], adaboost_num_estimators=[100], exp_name='post-index-fix')
# #util.feature_importances(results)



import multiprocessing
import numpy as np


def configure(num_jobs=8, TEST=False, subtract=0, num_proc=None, num_thread_per_proc=None):
    '''
    num_jobs is typically the # of genes we are parallelizing over
    '''
    if num_proc is None:
        num_proc = multiprocessing.cpu_count() - subtract

    if num_jobs > num_proc:
        num_jobs = num_proc

    if num_thread_per_proc is None:
        num_thread_per_proc = int(np.floor(num_proc/num_jobs))

    if TEST:
        num_jobs = 1
        num_thread_per_proc = 1

    try:
        import mkl
        mkl.set_num_threads(num_thread_per_proc)    
    except ImportError:
        print("MKL not available, so I'm not adjusting the number of threads")

    print("Launching %d jobs with %d MKL threads each" % (num_jobs, num_thread_per_proc))

    return num_jobs



import numpy as np
from ssk_cython import *

def weighted_degree_kxx(x, xp, d):
    assert len(x) == len(xp)
    l = len(x)
    sim = 0

    for k in range(1, d):
        beta_k = 2*(d-k+1)/float((d*(d+1)))
        sim_k = 0
        for i in range(1, l-k+1):
            if x[i:i+k] == xp[i:i+k]:
                sim_k += 1.
        sim += beta_k * sim_k

    return sim

def WD_K(sequences, d=4, cython=False):
    num_sequences = len(sequences)
    K = np.zeros((num_sequences, num_sequences))

    for i in range(num_sequences):
        for j in range(num_sequences):
            if cython:
                K[i, j] = cython_weighted_degree_kxx(sequences[i], sequences[j], d=d)
            else:
                K[i, j] = weighted_degree_kxx(sequences[i], sequences[j], d=d)

    return K

if __name__ == '__main__':
    x1 = 'ATCGATCG'
    x2 = 'ATCGATCG'
    x3 = 'ATCGATCC'
    x4 = 'ATCGATAA'
    x5 = 'ANNNNNNN'
    K = WD_K([x1, x2, x3, x4, x5])



"""
Module for setting memory thresholds cross-platform

Based on stackoverflow thread:
https://stackoverflow.com/questions/16779497/how-to-set-memory-limit-for-thread-or-process-in-python
"""
import os
import ctypes
import platform

PROCESS_SET_QUOTA = 0x100
PROCESS_TERMINATE = 0x1
JOB_OBJECT_LIMIT_PROCESS_MEMORY = 0x100
JOB_OBJECT_EXTENDED_LIMIT_INFORMATION = 9

class IO_COUNTERS(ctypes.Structure):
    _fields_ = [('ReadOperationCount', ctypes.c_uint64),
                ('WriteOperationCount', ctypes.c_uint64),
                ('OtherOperationCount', ctypes.c_uint64),
                ('ReadTransferCount', ctypes.c_uint64),
                ('WriteTransferCount', ctypes.c_uint64),
                ('OtherTransferCount', ctypes.c_uint64)]

class JOBOBJECT_BASIC_LIMIT_INFORMATION(ctypes.Structure):
    _fields_ = [('PerProcessUserTimeLimit', ctypes.c_int64),
                ('PerJobUserTimeLimit', ctypes.c_int64),
                ('LimitFlags', ctypes.c_uint32),
                ('MinimumWorkingSetSize', ctypes.c_void_p),
                ('MaximumWorkingSetSize', ctypes.c_void_p),
                ('ActiveProcessLimit', ctypes.c_uint32),
                ('Affinity', ctypes.c_void_p),
                ('PriorityClass', ctypes.c_uint32),
                ('SchedulingClass', ctypes.c_uint32)]

class JOBOBJECT_EXTENDED_LIMIT_INFORMATION(ctypes.Structure):
    _fields_ = [('BasicLimitInformation', JOBOBJECT_BASIC_LIMIT_INFORMATION),
                ('IoInfo', IO_COUNTERS),
                ('ProcessMemoryLimit', ctypes.c_void_p),
                ('JobMemoryLimit', ctypes.c_void_p),
                ('PeakProcessMemoryUsed', ctypes.c_void_p),
                ('PeakJobMemoryUsed', ctypes.c_void_p)]

def set_limit_windows(pid, size):
    """
    Set process memory limit on windowss

    Arguments:
        pid {int} -- Process ID
        size {int} -- Memory limit in bytes
    """
    job_info = JOBOBJECT_EXTENDED_LIMIT_INFORMATION()
    out_size = ctypes.c_uint32()

    job = ctypes.windll.kernel32.CreateJobObjectW(None, None)
    assert job != 0

    success = ctypes.windll.kernel32.QueryInformationJobObject(job,
                                                               JOB_OBJECT_EXTENDED_LIMIT_INFORMATION,
                                                               ctypes.POINTER(JOBOBJECT_EXTENDED_LIMIT_INFORMATION)(job_info),
                                                               ctypes.sizeof(JOBOBJECT_EXTENDED_LIMIT_INFORMATION),
                                                               ctypes.POINTER(ctypes.c_uint32)(out_size))
    assert success

    job_info.BasicLimitInformation.LimitFlags |= JOB_OBJECT_LIMIT_PROCESS_MEMORY
    job_info.ProcessMemoryLimit = size
    success = ctypes.windll.kernel32.SetInformationJobObject(job,
                                                             JOB_OBJECT_EXTENDED_LIMIT_INFORMATION,
                                                             ctypes.POINTER(JOBOBJECT_EXTENDED_LIMIT_INFORMATION)(job_info),
                                                             ctypes.sizeof(JOBOBJECT_EXTENDED_LIMIT_INFORMATION))
    assert success

    process = ctypes.windll.kernel32.OpenProcess(PROCESS_SET_QUOTA | PROCESS_TERMINATE,
                                                 False, pid)
    assert process != 0

    success = ctypes.windll.kernel32.AssignProcessToJobObject(job, process)
    assert success


    success = ctypes.windll.kernel32.CloseHandle(job)
    assert success

    success = ctypes.windll.kernel32.CloseHandle(process)
    assert success

def set_limit_unix(size):
    """
    Set process memory limit on windows

    Arguments:
        size {int} -- Memory limit in bytes
    """
    import resource
    soft, hard = resource.getrlimit(resource.RLIMIT_AS)
    resource.setrlimit(resource.RLIMIT_AS, (size, hard))

def set_limit(size):
    """
    Set process memory limit and throw a MemoryError when exceeded

        size {int} -- Memory limit in GiB
    """
    size_bytes = int(size * 1024 * 1024 * 1024)
    if platform.system() == "Windows":
        set_limit_windows(os.getpid(), size_bytes)
    else:
        set_limit_unix(size_bytes)



import numpy as np
import scipy as sp
import sklearn

def DNN_on_fold(feature_sets, train, test, y, y_all, X, dim, dimsum, learn_options):
    import theanets
    from sklearn.metrics import accuracy_score
    y = np.array(y_all[learn_options['DNN target variable']].values, dtype=float)
    y_train, X_train = y[train][:, None], X[train]
    y_test, X_test = y[test][:, None], X[test]

    num_hidden_layers = [1]#, 2, 3]
    num_units = [2]#, 5, 8, 10, 15, 20, 25, 30, 40, 50, 60]
    accuracies = np.zeros((len(num_hidden_layers), len(num_units)))
    best_score = None
    best_model = None

    for i, hl in enumerate(num_hidden_layers):
        for j, nu in enumerate(num_units):
            architecture = np.zeros((2+hl,))
            architecture[0] = X_train.shape[1]
            architecture[-1] = 1#len(np.unique(y_train))
            architecture[1:-1] = [nu for l in range(hl)]

            if learn_options["cv"] == "stratified":
                label_encoder = sklearn.preprocessing.LabelEncoder()
                label_encoder.fit(y_all['Target gene'].values[train])
                gene_classes = label_encoder.transform(y_all['Target gene'].values[train])
                n_folds = len(np.unique(gene_classes))
                cv = sklearn.cross_validation.StratifiedKFold(gene_classes, n_folds=n_folds, shuffle=True)
            elif learn_options["cv"]=="gene":
                gene_list = np.unique(y_all['Target gene'].values[train])
                cv = []
                for gene in gene_list:
                    cv.append(get_train_test(gene, y_all[train]))
                n_folds = len(cv)


            for train_ind, valid_ind in cv:
                # e = theanets.Experiment(
                #     theanets.Classifier,
                #     layers=architecture,
                #     train_batches=32,
                #     # patience=100,
                #     # tied_weights=False,
                #     )

                e = theanets.Experiment(
                    theanets.Regressor,
                    layers=architecture,
                    train_batches=32,
                    # patience=100,
                    # tied_weights=False,
                    )

                e.run((X_train[train_ind], y_train[train_ind]), (X_train[valid_ind], y_train[valid_ind]))
                pred = e.network.predict(X_train[valid_ind])

                accuracies[i, j] += sp.stats.spearmanr(pred.flatten(), y_train[valid_ind].flatten())[0]

            accuracies[i, j] = accuracies[i, j]/float(n_folds)


            if best_score is None or accuracies[i, j] > best_score:
                best_score = accuracies[i, j]
                best_model = copy.deepcopy(e)
                print("DNN with %d hidden layers and %d units, accuracy: %.4f   *" % (hl, nu, accuracies[i,j]))
            else:
                print("DNN with %d hidden layers and %d units, accuracy: %.4f" % (hl, nu, accuracies[i,j]))

    best_model.run((X_train, y_train), (X_test, y_test))
    y_pred = best_model.network.predict(X[test])

    return y_pred, None


"""
from https://gist.github.com/bwhite/3726239

Information Retrieval metrics

Useful Resources:
http://www.cs.utexas.edu/~mooney/ir-course/slides/Evaluation.ppt
http://www.nii.ac.jp/TechReports/05-014E.pdf
http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf
http://hal.archives-ouvertes.fr/docs/00/72/67/60/PDF/07-busa-fekete.pdf
Learning to Rank for Information Retrieval (Tie-Yan Liu)
"""
import numpy as np
import scipy as sp
import scipy.stats
from . import util as ut
import time

def mean_reciprocal_rank(rs):
    """Score is reciprocal of the rank of the first relevant item

    First element is 'rank 1'.  Relevance is binary (nonzero is relevant).

    Example from http://en.wikipedia.org/wiki/Mean_reciprocal_rank
    >>> rs = [[0, 0, 1], [0, 1, 0], [1, 0, 0]]
    >>> mean_reciprocal_rank(rs)
    0.61111111111111105
    >>> rs = np.array([[0, 0, 0], [0, 1, 0], [1, 0, 0]])
    >>> mean_reciprocal_rank(rs)
    0.5
    >>> rs = [[0, 0, 0, 1], [1, 0, 0], [1, 0, 0]]
    >>> mean_reciprocal_rank(rs)
    0.75

    Args:
        rs: Iterator of relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Mean reciprocal rank
    """
    rs = (np.asarray(r).nonzero()[0] for r in rs)
    return np.mean([1. / (r[0] + 1) if r.size else 0. for r in rs])


def r_precision(r):
    """Score is precision after all relevant documents have been retrieved

    Relevance is binary (nonzero is relevant).

    >>> r = [0, 0, 1]
    >>> r_precision(r)
    0.33333333333333331
    >>> r = [0, 1, 0]
    >>> r_precision(r)
    0.5
    >>> r = [1, 0, 0]
    >>> r_precision(r)
    1.0

    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        R Precision
    """
    r = np.asarray(r) != 0
    z = r.nonzero()[0]
    if not z.size:
        return 0.
    return np.mean(r[:z[-1] + 1])


def precision_at_k(r, k):
    """Score is precision @ k

    Relevance is binary (nonzero is relevant).

    >>> r = [0, 0, 1]
    >>> precision_at_k(r, 1)
    0.0
    >>> precision_at_k(r, 2)
    0.0
    >>> precision_at_k(r, 3)
    0.33333333333333331
    >>> precision_at_k(r, 4)
    Traceback (most recent call last):
        File "<stdin>", line 1, in ?
    ValueError: Relevance score length < k


    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Precision @ k

    Raises:
        ValueError: len(r) must be >= k
    """
    assert k >= 1
    r = np.asarray(r)[:k] != 0
    if r.size != k:
        raise ValueError('Relevance score length < k')
    return np.mean(r)


def average_precision(r):
    """Score is average precision (area under PR curve)

    Relevance is binary (nonzero is relevant).

    >>> r = [1, 1, 0, 1, 0, 1, 0, 0, 0, 1]
    >>> delta_r = 1. / sum(r)
    >>> sum([sum(r[:x + 1]) / (x + 1.) * delta_r for x, y in enumerate(r) if y])
    0.7833333333333333
    >>> average_precision(r)
    0.78333333333333333

    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Average precision
    """
    r = np.asarray(r) != 0
    out = [precision_at_k(r, k + 1) for k in range(r.size) if r[k]]
    if not out:
        return 0.
    return np.mean(out)


def mean_average_precision(rs):
    """Score is mean average precision

    Relevance is binary (nonzero is relevant).

    >>> rs = [[1, 1, 0, 1, 0, 1, 0, 0, 0, 1]]
    >>> mean_average_precision(rs)
    0.78333333333333333
    >>> rs = [[1, 1, 0, 1, 0, 1, 0, 0, 0, 1], [0]]
    >>> mean_average_precision(rs)
    0.39166666666666666

    Args:
        rs: Iterator of relevance scores (list or numpy) in rank order
            (first element is the first item)

    Returns:
        Mean average precision
    """
    return np.mean([average_precision(r) for r in rs])


def dcg_at_k(r, k, method=0):
    """Score is discounted cumulative gain (dcg)

    Relevance is positive real values.  Can use binary
    as the previous methods.

    Example from
    http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf
    >>> r = [3, 2, 3, 0, 0, 1, 2, 2, 3, 0]
    >>> dcg_at_k(r, 1)
    3.0
    >>> dcg_at_k(r, 1, method=1)
    3.0
    >>> dcg_at_k(r, 2)
    5.0
    >>> dcg_at_k(r, 2, method=1)
    4.2618595071429155
    >>> dcg_at_k(r, 10)
    9.6051177391888114
    >>> dcg_at_k(r, 11)
    9.6051177391888114

    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)
        k: Number of results to consider
        method: If 0 then weights are [1.0, 1.0, 0.6309, 0.5, 0.4307, ...]
                If 1 then weights are [1.0, 0.6309, 0.5, 0.4307, ...]

    Returns:
        Discounted cumulative gain
    """
    r = np.asfarray(r)[:k]
    if r.size:
        if method == 0:
            return r[0] + np.sum(r[1:] / np.log2(np.arange(2, r.size + 1)))
        elif method == 1:
            return np.sum(r / np.log2(np.arange(2, r.size + 2)))
        else:
            raise ValueError('method must be 0 or 1.')
    return 0.


def ndcg_at_k(r, k, method=0):
    """Score is normalized discounted cumulative gain (ndcg)

    Relevance is positive real values.  Can use binary
    as the previous methods.

    Example from
    http://www.stanford.edu/class/cs276/handouts/EvaluationNew-handout-6-per.pdf
    >>> r = [3, 2, 3, 0, 0, 1, 2, 2, 3, 0]
    >>> ndcg_at_k(r, 1)
    1.0
    >>> r = [2, 1, 2, 0]
    >>> ndcg_at_k(r, 4)
    0.9203032077642922
    >>> ndcg_at_k(r, 4, method=1)
    0.96519546960144276
    >>> ndcg_at_k([0], 1)
    0.0
    >>> ndcg_at_k([1], 2)
    1.0

    Args:
        r: Relevance scores (list or numpy) in rank order
            (first element is the first item)
        k: Number of results to consider
        method: If 0 then weights are [1.0, 1.0, 0.6309, 0.5, 0.4307, ...]
                If 1 then weights are [1.0, 0.6309, 0.5, 0.4307, ...]

    Returns:
        Normalized discounted cumulative gain
    """
    dcg_max = dcg_at_k(sorted(r, reverse=True), k, method)
    if not dcg_max:
        return 0.
    return dcg_at_k(r, k, method) / dcg_max

## ------------------------------------------------------------------------------------
## custom stuff from us to avoid problem with ties

def ndcg_at_k_ties(labels, predictions, k, method=0, normalize_from_below_too=False, theta=None):
    '''
    See 2008 McSherry et al on how to efficiently compute NDCG with ties
    labels are ground truth

    if k=None then k gets set to len(labels)

    labels and predictions get flattened here

    set normalize_from_below_too=False for conventional ndcg_at_k_ties, but note this will only
    ensure the max is 1, not that the min is zero. to get that added guarantee, set this argument to True
    '''

    if isinstance(labels, list):
        labels = np.array(labels)
    if isinstance(predictions, list):
        predictions = np.array(predictions)

            
    assert len(labels.shape)==1 or np.min(labels.shape)==1, "should be 1D array or equivalent"
    assert len(predictions.shape)==1 or np.min(predictions.shape)==1, "should be 1D array or equivalent"
        
    labels = labels.flatten()
    predictions = predictions.flatten()

    assert np.all(labels.shape == predictions.shape), "labels and predictions should have the same shape"
        
    if k is None:
        k = len(labels)

    labels = labels.copy()

    dcg = dcg_at_k_ties(labels, predictions, k, method=method, theta=theta)
        
    dcg_max = dcg_at_k_ties(labels, labels, k, method, theta=theta)
    # NOTE: I have checked that dcg_at_k_ties and dcg_at_k match when there are no ties, or ties in the labels

    
    if normalize_from_below_too:
        dcg_min = dcg_at_k_ties(np.sort(labels)[::-1], np.sort(predictions), k, method, theta=theta)
    else:
        dcg_min = 0
    numerator = (dcg - dcg_min)
    assert numerator > -1e-5
    numerator = np.max((0, numerator))
    ndcg = numerator / (dcg_max - dcg_min)            
    assert ndcg <= 1.0 and ndcg >= 0.0, "ndcg=%f should be in [0,1]" % ndcg
    if not dcg_max: 
        ndcg = 0.
    return ndcg

def dcg_helper(discount_factors, gain, k, labels, method, predictions):
    #step through, in current order (of decreasing predictions), accumulating tied gains (which may be singletons)
    ii = 0
    dcg = 0.0
    while (ii < k):
        current_pred = predictions[ii]
        current_gain = gain(labels[ii], method)
        # intializing the tied cumulative variables
        cum_tied_gain = current_gain
        cum_tied_disc = discount_factors[ii]
        num_ties = 1
        ii += 1
        # count number of ties in predictions
        while (ii<len(predictions) and predictions[ii]==current_pred):  #while tied
            num_ties += 1.0
            cum_tied_gain += gain(labels[ii], method)
            if ii < k: cum_tied_disc += discount_factors[ii]
            ii += 1
        #if len(np.unique(predictions))==1:  import ipdb; ipdb.set_trace()
        avg_gain = cum_tied_gain/num_ties
        dcg += avg_gain*cum_tied_disc
        assert not np.isnan(dcg), "found nan dcg"
    return dcg

def dcg_at_k_ties(labels, predictions, k, method=0, theta=None):
    '''
    See 2008 McSherry et al on how to efficiently compute NDCG (method=0 here) with ties (in the predictions)
    'labels' are what the "ground truth" judges assign
    'predictions' are the algorithm predictions corresponding to each label
    Also, http://en.wikipedia.org/wiki/Discounted_cumulative_gain for basic defns
    '''
    assert isinstance(predictions,np.ndarray)
    assert len(labels) == len(predictions), "labels and predictions should be of same length"
    assert k <= len(labels), "k should be <= len(labels)"

    # order both labels and preds so that they are in order of decreasing predictive score
    sorted_ind = np.argsort(predictions)[::-1]
    predictions = predictions[sorted_ind]
    labels = labels[sorted_ind]

    def gain(label, method):
        if method==0:
            return label
        elif method==1:
            return 2**label-1.0
        elif method==2 or method==3 or method==4:
            return label
        else:
            raise NotImplementedError()

    if method==0:
        discount_factors = get_discount_factors(len(labels), discount='log2')
    elif method==1:
        raise Exception("need to implement: log_2(i+1)")
    elif method==2:
        discount_factors = get_discount_factors(len(labels), discount='linear')
    elif method==3:
        discount_factors = get_discount_factors(len(labels), discount='combination')
    elif method==4:
        assert theta is not None, "need to specify theta or theta"
        discount_factors = get_discount_factors(len(labels), discount='1/rtheta', theta=theta)

    else:
        raise NotImplementedError()

    assert len(discount_factors) == len(labels), "discount factors has wrong length"

    dcg = dcg_helper(discount_factors, gain, k, labels, method, predictions)
    assert not np.isnan(dcg), "found nan dcg"
    
    return dcg

def get_discount_factors(num_labels, discount='log2', theta=None):
    ii_range = np.arange(num_labels) + 1

    if discount == 'log2':
        discount_factors = np.concatenate((np.array([1.0]), 1.0/np.log2(ii_range[1:])))
    elif discount == 'linear':
        discount_factors = -ii_range/float(num_labels) + 1.0
    elif discount == 'combination':
        l2 = np.concatenate((np.array([1.0]), 1.0/np.log2(ii_range[1:])))
        linear = -ii_range/float(num_labels) + 1.0
        discount_factors = np.max((l2,linear), axis=0)
    elif discount == '1/rtheta':
        discount_factors = 1./(ii_range**theta)
    else:
        raise NotImplementedError

    return discount_factors

def rank_data(r, rground):
    # we checked this heavily, and is correct, e.g. rground will go from largest rank to smallest
    r = sp.stats.mstats.rankdata(r)
    rground = sp.stats.mstats.rankdata(rground)
    assert np.sum(r)==np.sum(rground), "ranks should add up to the same"
    return r, rground

#todo: look at rank_data for bug, look at highest alpha predictions for ties
def dcg_alt(relevances, rank=20):
    relevances = np.asarray(relevances)[:rank]
    n_relevances = len(relevances)
    if n_relevances == 0:
        return 0.
    discounts = np.log2(np.arange(n_relevances) + 2)
    return np.sum(relevances / discounts)

def ndcg_alt(relevances, rank=20):
    best_dcg = dcg_alt(sorted(relevances, reverse=True), rank)
    if best_dcg == 0:
        return 0.
    return dcg_alt(relevances, rank) / best_dcg

def ndcg_bootstrap_test(preds1, preds2, true_labels, num_bootstrap, method, k, normalize_from_below_too, seed = "78923"):
    """
    Basic idea: use bootstrap to get the standard deviation of the difference in NDCG, and then create a z-statistic,
    z = (ndcg1-ndcg2)/std(ndcg1-ndcg2), and then compute a p-value under the assumption that this is normally distributed.
    Robin et al . BMC Bioinformatics 2011, 12:77
    http://www.biomedcentral.com/1471-2105/12/77
    """

    return pv

def ndcg_at_k_swap_perm_test(preds1, preds2, true_labels, nperm, method, k, normalize_from_below_too, theta=None, balance_zeros=True):
            
        # pVal is the probability that we would observe as big an AUC diff as we
        # did if the ROC curves were drawn from the null hypothesis (which is that 
        # one model does not perform better than the other)
        #
        # null hypothesis is that the prediction ranking are the same, so we exchange a random 
        # number of them with each other. 
        #
        # see ndcg_at_k_ties for all but the first four parameters
        #
        # balance_zeros = True means that when we swap a zero for a non-zero value, we will also do a reverse swap
        #
        # this is a two-sided test, but since it is a symmetric null distribution, one should
        # be able to divide the p-value by 2 to get the one-sided version (but think this through before using)
        
        if isinstance(preds1, list):
            preds1 = np.array(preds1)
        else:
            preds1 = preds1.flatten()

        if isinstance(preds2, list):
            preds2 = np.array(preds2)
        else:
            preds2 = preds2.flatten()

        if isinstance(true_labels, list):
            true_labels = np.array(true_labels)
        else:
            true_labels = true_labels.flatten()

        assert len(preds1) == len(preds2), "need same number of preditions from each model"
        assert len(preds1) == len(true_labels), "need same number of preditions in truth and predictions"
        N = len(preds1)

        # re-sort all by truth ordering so that when swap they are aligned
        sorted_ind = np.argsort(true_labels)[::-1]
        true_labels = true_labels[sorted_ind]
        preds1 = preds1[sorted_ind]
        preds2 = preds2[sorted_ind]

        ranks1 = sp.stats.mstats.rankdata(preds1)
        ranks2 = sp.stats.mstats.rankdata(preds2)
        
        ndcg1 = ndcg_at_k_ties(true_labels, ranks1, k=k, method=method, normalize_from_below_too=normalize_from_below_too, theta=theta)
        ndcg2 = ndcg_at_k_ties(true_labels, ranks2, k=k, method=method, normalize_from_below_too=normalize_from_below_too, theta=theta)

        real_ndcg_diff = {}
        perm_ndcg_diff = {}

        real_ndcg_diff = np.abs(ndcg1 - ndcg2)                
        perm_ndcg_diff = np.nan*np.zeros(nperm)
            
        if False:#np.all(preds1 == preds2):
            pval[theta] = 1.0            
        else:                    
            zero_ind = true_labels == 0
            assert np.sum(zero_ind) < len(zero_ind), "balancing assumes there are more zeros than ones"

            for t in range(nperm):
                pair_ind_to_swap = np.random.rand(N) < 0.5

                ranks1_perm = ranks1.copy();
                ranks1_perm[pair_ind_to_swap] = ranks2[pair_ind_to_swap]

                ranks2_perm = ranks2.copy();
                ranks2_perm[pair_ind_to_swap] = ranks1[pair_ind_to_swap]

                ndcg1_perm = ndcg_at_k_ties(true_labels, ranks1_perm, k=k, method=method, normalize_from_below_too=normalize_from_below_too, theta=theta)
                ndcg2_perm = ndcg_at_k_ties(true_labels, ranks2_perm, k=k, method=method, normalize_from_below_too=normalize_from_below_too, theta=theta)

                for theta in theta:
                    tmp_diff = np.abs(ndcg1_perm[theta] - ndcg2_perm[theta])
                    perm_ndcg_diff[theta][t] = tmp_diff

            pval = {}            
            
            num_stat_greater = np.max((((perm_ndcg_diff > real_ndcg_diff).sum() + 1), 1.0))
            pval = num_stat_greater / nperm

        if False:
            plt.figure();
            plt.plot(np.sort(perm_ndcg_diff), '.')
            plt.plot(real_ndcg_diff*np.ones(perm_ndcg_diff.shape), 'k-')
            plt.show()
                        
        return pval, real_ndcg_diff, perm_ndcg_diff, ndcg1, ndcg2

if __name__ == "__main__":
    import pickle as pickle
    import matplotlib.pyplot as plt
    import elevation.metrics
    from . import corrstats

    simulated_data = True
    permute_real_data = True
        
    T = 1000
    allp = np.nan*np.ones(T)

    nperm = 100
    #method = 4; normalize_from_below_too = True; 
    
    #theta_range = np.logspace(np.log10(0.01), np.log10(1.0), 3)  # Nicolo uses 10, so I grab the extremes and middle
    #theta_range = np.array([0.01])
    #weights = np.logspace(np.log10(0.0001), np.log10(10), 3); 
    #weights = np.array([100.0])
    weights = np.array([0.001])
    theta_range = weights# just to make life easier

    
    # only for simulated data
    N = 100
    frac_zeros = 0
    
    k = None

    allp = np.nan*np.zeros((len(theta_range) + 1, T))

    if not simulated_data:
        print("loading up saved data...") # two-fold CV data from CRISPR off-target GUIDE-SEQ
        with open(r'\\nerds5\kevin\from_nicolo\gs.pickle','rb') as f:  predictions, truth_all = pickle.load(f)
        print("done.")
        N = len(truth_all[0])
            
    for t in range(T):

        # totally simulated
        if simulated_data:
            truth = np.random.rand(N)
            zero_ind = np.random.rand(N) < frac_zeros
            truth[zero_ind] = 0
            pred1 = np.random.rand(N)
            pred2 = np.random.rand(N)
        else:                                
            fold = 0
            truth = truth_all[fold]
            pred1 = predictions["CFD"][fold]
            pred2 = predictions["product"][fold]
                        
            if permute_real_data:
                truth = np.random.permutation(truth)

        t0 = time.time()        
        #pval, real_ndcg_diff,  perm_ndcg_diff, ndcg1, ndcg2 = ndcg_at_k_swap_perm_test(pred1, pred2, truth, nperm, method, k, normalize_from_below_too, theta_range=theta_range)        
        for i, w in enumerate(weights):
            weights_array = truth.copy()
            weights_array += w

            #corr0 = elevation.metrics.spearman_weighted(truth, pred1, w=weights_array)
            #corr1 = elevation.metrics.spearman_weighted(truth, pred2, w=weights_array)
            #corr01 = elevation.metrics.spearman_weighted(pred1, pred2, w=weights_array)
            #n0 = len(truth)        
            #t2, pvaltmp = corrstats.dependent_corr(corr0, corr1, corr01, n0, twotailed=True, method="steiger")

            pvaltmp, real_corr_diff, perm_corr_diff, corr1, corr2 = elevation.spearman_weighted_swap_perm_test(pred1, pred2, truth, nperm, weights_array)
                                                        
            allp[i, t] = pvaltmp
            t1 = time.time()

        #for i, theta in enumerate(theta_range.tolist() + ["all"]):
        #    print "%d, theta=%s) ndcg1=%f, ndcg2=%f, ndcg_diff=%f, p=%f, elapsed time=%f minutes, smallest_p=%f" % (t, str(theta), ndcg1[theta], ndcg2[theta], real_ndcg_diff[theta], pval[theta], (t1-t0)/60, 1.0/nperm)        
        #    allp[i, t] = pval[theta]
        #print "---------------"
        
    #for i, theta in enumerate(theta_range.tolist() + ["all"]):
    for i, theta in enumerate(theta_range.tolist()):
        #mytitle = "Norm. hist p-values nDCG\n %d null samples, w %d perm and N=%d, theta=%s" % (T, nperm, N, str(theta))
        mytitle = "Norm. hist p-values Steiger w weighted Spearman\n %d null samples, N=%d, weight=%s" % (T, N, str(theta))
        ut.qqplotp(allp[i,:], dohist=True, numbins=10, figsize=[6,6], title=mytitle, markersize=5)
        plt.show()
    
    #save_tmp_results = r'D:\Source\CRISPR\elevation\pickles\tmp.ndcg.stat.calibration.p'
    #pickle.dump([theta_range, allp, pval, real_ndcg_diff, perm_ndcg_diff, ndcg1, ndcg2], open(save_tmp_results, "wb" ))
    #[theta_range, allp, pval, real_ndcg_diff, perm_ndcg_diff, ndcg1, ndcg2] = pickle.load(open(save_tmp_results, "rb" ))


    import ipdb; ipdb.set_trace()


    # # e.g. where all predictions are the same
    # labels = np.arange(30)
    # predictions = np.ones(30)

    # discount_factors = get_discount_factors(labels)
    # avg_label = np.mean(labels)
    # avg_label_vec = avg_label*np.ones((len(labels),1))

    # for k in range(10):
    #     # one way
    #     dcg1 = np.dot(discount_factors[0:k,None].T, avg_label_vec[0:k])[0][0]
    #     # another way
    #     dcg2 = np.sum(discount_factors[0:k])*avg_label
    #     # using our function
    #     dcg3 = dcg_at_k_ties(labels,predictions,k)

    #     print "%f, %f, %f" % (dcg1, dcg2, dcg3)
    #     assert(np.abs(dcg1 - dcg2) < 1e-8)
    #     assert(np.abs(dcg2 - dcg3) < 1e-8)
    # print "check out ok for case with all ties in predictions"

    truth = np.array([3, 4, 2, 1, 0, 0, 0])
    pred1 = np.array([3, 4, 2, 1, 0, 0, 0])
    pred2 = np.array([2, 1, 3, 4, 5, 6, 7])

    truth3 = np.array([3, 4, 2, 1, 0, 0, 0])
    truth4 = np.zeros(7); truth4[0] = 1
    pred3 = np.array([2, 1, 3, 4, 5, 6, 7])*10
    pred4 = np.array([4, 3, 2, 1, 0, 0, 0])
    pred5 = np.array([4, 3, 1, 2, 0, 0, 0])

    nperm = 1000
    method = 4; theta = 0.5; normalize_from_below_too = True
    k = len(pred3)

    pval, real_ndcg_diff,  perm_ndcg_diff, ndcg1, ndcg2 = ndcg_at_k_swap_perm_test(pred1, pred2, truth, nperm, method, k, normalize_from_below_too, theta=theta)
    print("ndcg1=%f, ndcg2=%f, ndcg_diff=%f, p=%f" % (ndcg1, ndcg2, real_ndcg_diff, pval))
    
    pval, real_ndcg_diff,  perm_ndcg_diff, ndcg1, ndcg2 = ndcg_at_k_swap_perm_test(pred1, pred1, truth, nperm, method, k, normalize_from_below_too, theta=theta)    
    print("ndcg1=%f, ndcg2=%f, ndcg_diff=%f, p=%f" % (ndcg1, ndcg2, real_ndcg_diff, pval))

    pval, real_ndcg_diff,  perm_ndcg_diff, ndcg1, ndcg2 = ndcg_at_k_swap_perm_test(pred1, pred4, truth, nperm, method, k, normalize_from_below_too, theta=theta)    
    print("ndcg1=%f, ndcg2=%f, ndcg_diff=%f, p=%f" % (ndcg1, ndcg2, real_ndcg_diff, pval))

    pval, real_ndcg_diff,  perm_ndcg_diff, ndcg1, ndcg2 = ndcg_at_k_swap_perm_test(pred1, pred5, truth, nperm, method, k, normalize_from_below_too, theta=theta)    
    print("ndcg1=%f, ndcg2=%f, ndcg_diff=%f, p=%f" % (ndcg1, ndcg2, real_ndcg_diff, pval))

    import ipdb; ipdb.set_trace()


    #print ndcg_at_k_ties(truth, truth, k, method=0, normalize_from_below_too=True)
    #print ndcg_at_k_ties(truth, pred2, k, method=0, normalize_from_below_too=True)
    #print ndcg_at_k_ties(truth, pred3, k, method=0, normalize_from_below_too=True)
    #print ndcg_at_k_ties(truth3, pred3, k, method=3, normalize_from_below_too=True)
    print(ndcg_at_k_ties(truth4, pred2, k, method=3, normalize_from_below_too=True))
        
    print(ndcg_alt(truth[np.argsort(pred2)[::-1]], 5))
    print(ndcg_at_k(truth[np.argsort(pred2)[::-1]], 5, method=1))
    print(ndcg_at_k(truth[np.argsort(pred2)[::-1]], 5, method=0))

    print(ndcg_at_k_ties(truth, pred2, 5, method=1))
    print(ndcg_at_k_ties(truth, pred2, 5, method=0))



import numpy as np
import sklearn
from sklearn.svm import LinearSVC
from sklearn.linear_model import LogisticRegression
import sklearn.linear_model
import pandas

def mean_on_fold(feature_sets, train, test, y, y_all, inputs, dim, dimsum, learn_options):
    return np.ones((test.sum(), 1))*y[train].mean(), None


def random_on_fold(feature_sets, train, test, y, y_all, inputs, dim, dimsum, learn_options):
    return np.random.randn(test.sum(), 1), None


def xu_et_al_on_fold(feature_sets, train, test, y, y_all, X, dim, dimsum, learn_options):
    coef = pandas.read_csv(learn_options['xu_matrix_file'], skiprows=1, delimiter='\t')
    coef = coef[['A', 'T', 'C', 'G']] # swap columns so that they are in correct order
    coef = coef.values.flatten()[:, None]
    X = X.copy()
    X = np.append(X, np.zeros((X.shape[0], 3*4)), axis=1)
    X = X[:, 3*4:]
    y_pred = 1./(1+np.exp(-np.dot(X[test], coef)))

    return y_pred, coef

def doench_on_fold(feature_sets, train, test, y, y_all, X, dim, dimsum, learn_options):
    auto_class_weight = None  # 'auto'/None
    verbose = False
    penalty = [0.005*pow(1.15, x) for x in range(0, 45)]  # ian's code:  tvals = [0.005*pow(1.15,x) for x in range(0,45)]
    y_bin = y_all[learn_options['binary target name']].values[:, None]

    label_encoder = sklearn.preprocessing.LabelEncoder()
    label_encoder.fit(y_all['Target gene'].values[train])
    gene_classes = label_encoder.transform(y_all['Target gene'].values[train])
    cv = sklearn.cross_validation.StratifiedKFold(gene_classes, n_folds=10, shuffle=True)

    best_penalty = None

    cv_results = np.zeros((10, len(penalty)))

    for j, split in enumerate(cv):
        train_inner, test_inner = split
        for i, c in enumerate(penalty):
            # fit an L1-penalized SVM classifier
            clf = LinearSVC(penalty='l1', C=c, dual=False, class_weight=auto_class_weight)
            clf.fit(X[train][train_inner], y_bin[train][train_inner].flatten())

            # pass features with non-zero coeff to Logistic with l2 penalty (original code?)
            non_zero_coeff = (clf.coef_ != 0.0)

            if np.all(non_zero_coeff is False):
                # if all are zero, turn one on so as to be able to run the code.
                non_zero_coeff[0] = True

            clf = LogisticRegression(penalty='l2', class_weight=auto_class_weight)
            clf.fit(X[train][train_inner][:, non_zero_coeff.flatten()], y[train][train_inner].flatten())
            y_test = clf.predict_proba(X[train][test_inner][:, non_zero_coeff.flatten()])[:, 1]

            fpr, tpr, _ = sklearn.metrics.roc_curve(y_bin[train][test_inner], y_test)
            assert np.nan not in fpr, "found nan fpr"
            assert np.nan not in tpr, "found nan tpr"
            roc_auc = sklearn.metrics.auc(fpr, tpr)
            if verbose:
                print(j, i, roc_auc)
            cv_results[j][i] = roc_auc

    best_penalty = penalty[np.argmax(np.mean(cv_results, axis=0))]
    print("best AUC for penalty: ", np.median(cv_results, axis=0))
    clf = LinearSVC(penalty='l1', C=best_penalty, dual=False, class_weight=auto_class_weight)
    clf.fit(X[train], y_bin[train].flatten())
    non_zero_coeff = (clf.coef_ != 0.0)

    clf = LogisticRegression(penalty='l2', class_weight=auto_class_weight)
    clf.fit(X[train][:, non_zero_coeff.flatten()], y[train].flatten())
    y_pred = clf.predict_proba(X[test][:, non_zero_coeff.flatten()])[:, 1:2]

    return y_pred, clf


def sgrna_from_doench_on_fold(feature_sets, train, test, y, y_all, X, dim, dimsum, learn_options):
    assert len(list(feature_sets.keys())) == 1, "should only use sgRNA Score here"
    assert list(feature_sets.keys())[0] == "sgRNA Score"
    y_pred = X[test][:, 0]
    return y_pred, None


def SVC_on_fold(feature_sets, train, test, y, y_all, X, dim, dimsum, learn_options):
    y_bin = y_all[learn_options['binary target name']].values[:, None]
    clf = LinearSVC(penalty='l2', dual=False)
    clf.fit(X[train], y_bin[train].flatten())
    #y_pred = clf.predict(X[test])[:, None] # this returns 0/1
    y_pred = clf.decision_function(X[test])[:, None]
    return y_pred, clf


    



"""
Module for calculating off target binding effects
"""
import math
import time
import logging
from multiprocessing.pool import ThreadPool
import numpy as np
import pandas as pd
from optimal_guide_finder.cas_model import CasModel

NT_POS = {'A':0, 'T':1, 'C':2, 'G':3}

def initalize_model(filename):
    """
    Init Cas9 binding model

    Args:
        filename (str): Path to genome file generated using get_sequence

    Returns:
        CasModel
    """
    logger = logging.getLogger(__name__)

    __start = time.time()
    logger.info("Creating Model...")
    model = CasModel(filename)
    __elasped = (time.time() - __start)
    logger.info("Time Model Building: {:.2f}".format(__elasped))

    return model

def process_guides(model, guide_info, num_threads=None):
    """
    Process guides for off target binding effects using model for Cas9 binding

    Args:
        model (CasModel): Initialized model object
        guide_info (dict): Dictionary gernerated by guide generator containing guides information
        num_threads (int, optional): max number of threads to use, None will use all available threads.
                                     Defaults to None

    Returns:
        Dataframe: Pandas dataframe containing results
    """
    logger = logging.getLogger(__name__)

    info_df = pd.DataFrame()
    logger.info("Processing Guides...")
    __start = time.time()

    pool = ThreadPool(processes=num_threads)
    results = []
    for gene in guide_info:
        for i, guide in enumerate(guide_info[gene][0]):
            guide_data = pd.Series([guide, gene, guide_info[gene][1][i], guide_info[gene][2][i]],
                                   index=["Guide Sequence", "Gene/ORF Name", "Location in Gene", "Strand"])

            # call
            res = pool.apply_async(_process_guide, (model, guide, logger))
            results.append(res)
            info_df = info_df.append(guide_data, ignore_index=True)

    pool.close()

    # pull results from the queue
    result_df = pd.DataFrame()
    for res in results:
        result_df = result_df.append(res.get(), ignore_index=True)

    results_df = pd.merge(info_df, result_df, on='Guide Sequence')

    __elasped = (time.time() - __start)
    logger.info("Time Spent Analysing Guides: {:.2f}".format(__elasped))

    return results_df

def _process_guide(model, guide, logger):
    """
    Process a single guide for off target binding effects

    Args:
        model (CasModel): initialized model
        guide (string): guide sequence
        logger (logging.logger): logger instance to use

    Returns:
        [Series]: Pandas Series containing results for guide
    """
    num_guide = np.array([NT_POS[nt] for nt in list(guide)])
    partition_function = 1

    result = []
    dgs = []

    for (source, _) in model.genome_dictionary.items():
        for full_pam in model.get_all_pams():
            dg_pam = model.calc_dg_pam(full_pam)
            dg_supercoiling = model.calc_dg_supercoiling(sigma_initial=-0.05, target_seq=20 * "N")

            for (target_sequence, _) in model.genome_dictionary[source][full_pam]:
                np_target_sequence = np.array([NT_POS[nt] for nt in list(target_sequence)])
                dg_exchange = model.calc_dg_exchange(num_guide, np_target_sequence)
                dg_target = dg_pam + dg_supercoiling + dg_exchange

                result.append([target_sequence, math.exp(-dg_target / model.RT)])
                partition_function += math.exp(-dg_target / model.RT)
                dgs.append(dg_target)

    result.insert(0, [guide, partition_function])
    guide_series = _process_off_target_guides(result)
    logger.info(guide_series)
    dgs.sort()
    info_dict = _info_logging(partition_function, dgs[:5], model.RT)
    info_dict['guide'] = guide
    logger.info(info_dict)

    return guide_series

def _process_off_target_guides(guide_result):
    """
    Post processing on guide result to calculate entropy and summarize results

    Args:
        guide_result (list): result from processing guides

    Returns:
        [Series]: Pandas series summarizing results for guide
    """
    guide_seq = guide_result[0][0]
    partition_function = guide_result[0][1]
    guide_entropy = 0
    exact_matches = 0
    for off_target in guide_result[1:]:
        if guide_seq == off_target[0]:
            exact_matches += 1
        probability = off_target[1]/partition_function
        guide_entropy -= probability*np.log2(probability)

    guide_series = pd.Series([guide_seq,
                              guide_entropy,
                              exact_matches],
                             index=["Guide Sequence",
                                    "Entropy Score",
                                    "Number of Exact Matches"])
    return guide_series

def _info_logging(partition_function, dgs, RT):
    """
    Summarizes information into dictionary for logging
    """
    info_dict = {'dg':[], 'probability':[]}

    for dg in dgs:
        info_dict['dg'].append(dg)
        info_dict['probability'].append(math.exp(-dg/RT)/partition_function)

    return info_dict



"""
Module responsible for calculating the per guide binding
"""
import os
import logging
import numpy as np
import scipy.io
from Bio import SeqIO
from numba import jit
from tqdm import tqdm

class CasModel():
    """
    Model for Cas9 binding based on biophysical properties of DNA binding
    """
    RT = 0.61597

    # the PAMs with the highest dG, ignoring other PAM sequences by setting their dG to 0
    _PAM_ENERGY = {
        'GGA': -9.8, 'GGT': -10, 'GGC': -10, 'GGG': -9.9, 'CGG': -8.1, 'TGG': -7.8, 'AGG': -8.1,
        'AGC': -8.1, 'AGT': -8.1, 'AGA': -7.9, 'GCT': -7.1, 'GCG': -6.9, 'ATT': -7.2, 'ATC': -6.4,
        'TTT': -7.6, 'TTG': -6.8, 'GTA': -7.4, 'GTT': -7.9, 'GTG': -7.7, 'AAT': -7, 'AAG': -7,
        'TAT': -7.2, 'TAG': -7.2, 'GAA': -7.2, 'GAT': -7.3, 'GAC': -7.2, 'GAG': -7.3
    }

    def __init__(self, filename, model_name='data/InvitroModel.mat'):
        """
        Initialize a CasModel object

        Arguments:
            filename {string} -- location of saved genome and target

        Keyword Arguments:
            model_name {str} -- path to mat file containing model data (default: {'data/InvitroModel.mat'})
        """
        root_path = os.path.dirname(__file__)
        model_name = os.path.join(root_path, model_name)
        self._logger = logging.getLogger(__name__)
        self._model_name = model_name
        self._nt_mismatch_in_first8_list = []

        data = scipy.io.loadmat(self._model_name)

        self._weights = data['w1']
        self._new_weights = np.array([weight[0]*2.3 for weight in self._weights])
        self._dec_nn = data['decNN']

        self._init_genome_finder(filename)

    def print_model_info(self):
        """
        Prints current model information
        """
        count = 0
        sums = 0
        negative_val = 0
        for i, mismatch in enumerate(self._dec_nn):
            for j, energy in enumerate(mismatch):
                if float(energy) < 0:
                    negative_val += 1
                if i != j:
                    sums += float(energy)
                    count += 1
        mean_nn = float(sums)/float(count)

        sum_weights = 0
        for weight in self._weights:
            sum_weights += weight
        meanw = sum_weights/len(self._weights)

        self._logger.info("Average mismatch energy: %d", mean_nn)
        self._logger.info("Average weight: %d", meanw)
        self._logger.info("Number of negative energies: %d", negative_val)

    def get_all_pams(self):
        """
        Generates a list of all possible PAMs

        Yields:
            [string]: Pam seqs
        """
        # PAM part will be 'GGT'
        for (pam_part, _) in sorted(list(self._PAM_ENERGY.items()), key=lambda x: x[1]):
            for nucleotide in ('A', 'G', 'C', 'T'):  # nt + PAMpart will be all possible 'NGGT'
                yield nucleotide + pam_part

    def calc_dg_pam(self, pam_full_seq):
        """
        Calculates the delta-G for full pam sequence
        """
        # PAM sequence is 5' - N xxx N - 3' where the energy of xxx is listed below.
        # a normal PAM of 'NGG' with 'TC' afterwards would be listed as 'GGT'
        key = pam_full_seq[1:4]
        if key in self._PAM_ENERGY:
            return self._PAM_ENERGY[key]

        return 0.0

    def calc_dg_exchange(self, guide_seq, target_seq):
        """
        Calculate the Delta-G Exchange for guide and target sequences
        """
        self._nt_mismatch_in_first8_list = []

        solverfunc = self._quick_calc_exchange_energy
        dg_exchange = solverfunc(self._new_weights, guide_seq, target_seq)

        return dg_exchange

    def calc_dg_supercoiling(self, sigma_initial, target_seq):
        """
        Calculate the Delta-G supercoiling for guide and target sequences
        """
        sigma_final = -0.08
        dg_supercoiling = 10.0 * \
            len(target_seq) * self.RT * (sigma_final**2 - sigma_initial**2)
        return dg_supercoiling

    def _init_genome_finder(self, filename):
        """
        Initializes genome dictionary by finding all potrntial locations in genome for off-target binding

        Args:
            filename (string): Path to genome and target files
        """
        genome_dictionary = {}

        handle = open(filename, 'r')
        records = SeqIO.parse(handle, "fasta")
        record = next(records)
        handle.close()

        full_sequence = str(record.seq)

        positions_at_mers = self._identify_nucleotide_positions_of_mers(full_sequence, 10)

        self._logger.info("Identifying target sites...")
        genome_dictionary[filename] = {}
        target_sequence_list = []
        mers_omitted_tally = 0

        for full_pam in tqdm(self.get_all_pams()):
            target_list_output = \
                self._identify_target_sequences_matching_pam(full_pam, positions_at_mers, full_sequence)
            target_sequence_list = target_list_output[0]
            mers_omitted_tally += target_list_output[1]
            genome_dictionary[filename][full_pam] = target_sequence_list

        if mers_omitted_tally > 0:
            self._logger.error('%d k-mers with non-nucleotide characters omitted from target list', mers_omitted_tally)

        self.genome_dictionary = genome_dictionary

    @staticmethod
    @jit(nopython=True)
    def _quick_calc_exchange_energy(weights, cr_rna, target_seq):
        """
        Calculate the delta G value a potential guide against the target sequence
        this method is made static to allow compilation with numba for speed-up
        """
        dg = 0
        for i in range(len(cr_rna)):
            pos = 20 - i
            if cr_rna[i] == target_seq[i]:
                continue
            else:
                dg += weights[pos]
        return float(dg)

    def _mers(self, length):
        """
        Generates multimers for sorting through list of 10mers based on user
        specification. Multimers generated act as the keys for generating a
        hashtable to eliminate undesired sequence patterns from those 10mers not
        found in the genome.

        Usage: mers(N) = 4^(N) unique Nmers
        """
        seq_list = ['']
        counter = 0

        while counter < length:
            for seq in seq_list:
                if len(seq) == counter:
                    for x in ['A', 'T', 'C', 'G']:
                        seq_list.append(seq + x)
            counter += 1

        last_n_mers = 4**length
        return seq_list[len(seq_list)-last_n_mers:]

    def _identify_nucleotide_positions_of_mers(self, full_sequence, length=10):
        """
        Saves list of nucleotide positions in genome that all match a unique N-mer
        sequence. Counting begins at __ending_ of MER.

        Usage:  genomePositionsAtMers[mer_sequence] is a list of nucleotide positions
                within the inputted fullSequence that match mer_sequence
                If mer_sequence ends in a PAM site, then this can be used to match
                the first N-3 nt of a guide strand plus a PAM site sequence.
        """
        # Create a list of all possible N-mers
        all_possible_mers = self._mers(length)

        # Search through the genome and add nucleotide positions for match to an N-mer
        positions_at_mers = {}
        for mer in all_possible_mers:
            positions_at_mers[mer] = []

        counter = 0

        non_nucleotide_count = 0 # Added
        while counter < (len(full_sequence)-length):
            word = full_sequence[counter: counter+length]
            try:
                positions_at_mers[word].append(counter+length)
            except KeyError:
                non_nucleotide_count += 1
            counter += 1

        if non_nucleotide_count > 0:
            self._logger.error('%d k-mers with non-nucleotide characters identified', non_nucleotide_count)

        return positions_at_mers

    def _identify_target_sequences_matching_pam(self, pam_seq, positions_at_mers, full_sequence,
                                                target_sequence_length=20):
        """
        Generates a list of target nucleotide sequences and corresponding nt
        positions for an inputted sequence that matched the PAM_seq.
                Uses the positionsAtMers dictionary to accelerate the identification.
                Good for large genomes.

        Usage:  listOfTargets = identifyTargetSequencesMatchingPAM('CGG',
                                                                positionsAtMers, genome_sequence)
        """
        target_sequence_list = []
        all_mers = list(positions_at_mers.keys())
        mer_length = len(all_mers[0])
        mers_omitted_count = 0
        list_of_mers_with_pam = [
            mer + pam_seq for mer in self._mers(mer_length - len(pam_seq))]
        for mer_with_pam in list_of_mers_with_pam:
            nt_list = positions_at_mers[mer_with_pam]
            for nt in nt_list:
                begin = nt-target_sequence_length - len(pam_seq)
                end = nt - len(pam_seq)
                # Does not account for circular DNAs
                if begin > 0 and end < len(full_sequence):
                    target_sequence = full_sequence[begin: end]
                    if set(target_sequence).issubset('ATCG'):
                        target_sequence_list.append((target_sequence, nt))
                        # print(target_sequence, ' : ', nt)
                    else:
                        mers_omitted_count += 1
        # print(mers_omitted_count)
        target_list_output = [target_sequence_list, mers_omitted_count]
        return target_list_output



import pandas
from . import util
import matplotlib.pyplot as plt
import scipy as sp
import scipy.stats
import numpy as np
import os

cur_dir = os.path.dirname(os.path.abspath(__file__))

def from_custom_file(data_file, learn_options):
    # use semantics of when we load V2 data
    print("Loading inputs to predict from %s" % data_file)
    data = pandas.read_csv(data_file)

    mandatory_columns = ['30mer', 'Target gene', 'Percent Peptide', 'Amino Acid Cut position']
    for col in mandatory_columns:
        assert col in data.columns, "inputs for prediction must include these columns: %s" % mandatory_columns

    Xdf = pandas.DataFrame(data)
    Xdf['30mercopy'] = Xdf['30mer']
    Xdf = Xdf.set_index(['30mer', 'Target gene'])
    Xdf['30mer'] = Xdf['30mercopy']
    Xdf.index.names = ['Sequence', 'Target']
    Xdf['drug']= ['dummydrug%s' % i for i in range(Xdf.shape[0])]
    Xdf = Xdf.set_index('drug', append=True)

    Y = None
    gene_position = Xdf[['Percent Peptide', 'Amino Acid Cut position']]
    target_genes =  np.unique(Xdf.index.levels[1])

    learn_options = set_V2_target_names(learn_options)

    return Xdf, Y, gene_position, target_genes


def from_file(data_file, learn_options, data_file2=None, data_file3=None):
    if learn_options["V"] == 1:  # from Nature Biotech paper

        print("loading V%d data" % learn_options["V"])

        assert not learn_options["weighted"] is not None, "not supported for V1 data"
        annotations, gene_position, target_genes, Xdf, Y = read_V1_data(data_file, learn_options)

        learn_options['binary target name'] = 'average threshold'
        learn_options['rank-transformed target name'] = 'average rank'
        learn_options['raw target name'] = 'average activity'

        # NF: not sure why the line below was uncommented
        # gene_position, selected_ind, target_genes, Xdf, Y = extract_by_organism("mouse", Xdf, Y, gene_position)

    elif learn_options["V"] == 2:  # from Nov 2014, hot off the machines
        Xdf, drugs_to_genes, target_genes, Y, gene_position = read_V2_data(data_file, learn_options)

        # check that data is consistent with sgRNA score
        xx = Xdf['sgRNA Score'].values
        yy = Y['score_drug_gene_rank'].values
        rr,pp = sp.stats.pearsonr(xx, yy)
        assert rr > 0, "data processing has gone wrong as correlation with previous predictions is negative"

        learn_options = set_V2_target_names(learn_options)

    elif learn_options["V"] == 3:  # merge of V1 and V2--this is what is used for the final model
        # these are relative to the V2 data, and V1 will be made to automatically match
        learn_options['binary target name'] = 'score_drug_gene_threshold'
        learn_options['rank-transformed target name'] = 'score_drug_gene_rank'
        learn_options['raw target name'] = None

        Xdf, Y, gene_position, target_genes = mergeV1_V2(data_file, data_file2, learn_options)


    elif learn_options["V"] == 4:  # merge of V1 and V2 and the Xu et al data
        # these are relative to the V2 data, and V1 and Xu et al. will be made to automatically match
        learn_options['binary target name'] = 'score_drug_gene_threshold'
        learn_options['rank-transformed target name'] = 'score_drug_gene_rank'
        learn_options['raw target name'] = None

        Xdf, Y, gene_position, target_genes = merge_all(data_file, data_file2, data_file3, learn_options)


    elif learn_options['V'] == 5:
        learn_options['binary target name'] = 'score_drug_gene_threshold'
        learn_options['rank-transformed target name'] = 'score_drug_gene_rank'
        learn_options['raw target name'] = None

        gene_position, target_genes, Xdf, Y = read_xu_et_al(data_file3)


    # truncate down to 30--some data sets gave us more.
    Xdf["30mer"] = Xdf["30mer"].apply(lambda x: x[0:30])

    return Xdf, Y, gene_position, target_genes


def set_V2_target_names(learn_options):
    if 'binary target name' not in list(learn_options.keys()):
        learn_options['binary target name'] = 'score_drug_gene_threshold'
    if 'rank-transformed target name' not in list(learn_options.keys()):
        learn_options['rank-transformed target name'] = 'score_drug_gene_rank'
    learn_options['raw target name'] = 'score'
    return learn_options


def combine_organisms(human_data, mouse_data):
    # 'Target' is the column name, 'CD13' are some rows in that column
    # xs slices through the pandas data frame to return another one
    cd13 = human_data.xs('CD13', level='Target', drop_level=False)
    # y_names are column names, cd13 is a pandas object
    X_CD13, Y_CD13 = util.get_data(cd13, y_names=['NB4 CD13', 'TF1 CD13'])
    cd33 = human_data.xs('CD33', level='Target', drop_level=False)
    X_CD33, Y_CD33 = util.get_data(cd33, y_names=['MOLM13 CD33', 'TF1 CD33', 'NB4 CD33'])
    cd15 = human_data.xs('CD15', level='Target', drop_level=False)
    X_CD15, Y_CD15 = util.get_data(cd15, y_names=['MOLM13 CD15'])

    mouse_X = pandas.DataFrame()
    mouse_Y = pandas.DataFrame()
    for k in mouse_data.index.levels[1]:
        # is k the gene
        X, Y = util.get_data(mouse_data.xs(k, level='Target', drop_level=False), ["On-target Gene"], target_gene=k, organism='mouse')
        mouse_X = pandas.concat([mouse_X, X], axis=0)
        mouse_Y = pandas.concat([mouse_Y, Y], axis=0)

    X = pandas.concat([X_CD13, X_CD15, X_CD33, mouse_X], axis=0)
    Y = pandas.concat([Y_CD13, Y_CD15, Y_CD33, mouse_Y], axis=0)

    return X, Y


def read_V1_data(data_file, learn_options, AML_file=cur_dir + "/data/V1_suppl_data.txt"):
    if data_file is None:
        data_file = cur_dir + "/data/V1_data.xlsx"
    human_data = pandas.read_excel(data_file, sheetname=0, index_col=[0, 1])
    mouse_data = pandas.read_excel(data_file, sheetname=1, index_col=[0, 1])
    Xdf, Y = combine_organisms(human_data, mouse_data)

    # get position within each gene, then join and re-order
    # note that 11 missing guides we were told to ignore
    annotations = pandas.read_csv(AML_file, delimiter='\t', index_col=[0, 4])
    annotations.index.names = Xdf.index.names
    gene_position = pandas.merge(Xdf, annotations, how="inner", left_index=True, right_index=True)
    gene_position = util.impute_gene_position(gene_position)
    gene_position = gene_position[['Amino Acid Cut position', 'Nucleotide cut position', 'Percent Peptide']]
    Y = Y.loc[gene_position.index]
    Xdf = Xdf.loc[gene_position.index]

    Y['test'] = 1  # for bookeeping to keep consistent with V2 which uses this for "extra pairs"

    target_genes = Y['Target gene'].unique()

    Y.index.names = ['Sequence', 'Target gene']

    assert Xdf.index.equals(Y.index), "The index of Xdf is different from the index of Y (this can cause inconsistencies/random performance later on)"

    if learn_options is not None and learn_options["flipV1target"]:
        print("************************************************************************")
        print("*****************MATCHING DOENCH CODE (DEBUG MODE)**********************")
        print("************************************************************************")
        # normally it is: Y['average threshold'] = Y['average rank'] > 0.8, where 1s are good guides, 0s are not
        Y['average threshold'] = Y['average rank'] < 0.2  # 1s are bad guides
        print("press c to continue")
        import ipdb
        ipdb.set_trace()

    return annotations, gene_position, target_genes, Xdf, Y

def rank_transform(x):
    return 1.0 - sp.stats.mstats.rankdata(x)/sp.stats.mstats.rankdata(x).max()

def read_xu_et_al(data_file, learn_options=None, verbose=True, subsetting='ours'):
    if data_file is None:
        data_file = '../data/xu_et_al_data.xlsx'

    datasets = ['ribo', 'non_ribo', 'mESC']
    aggregated = None

    for d in datasets:
        data_efficient = pandas.read_excel(data_file, sheetname='%s_efficient_sgRNA' % d, skiprows=2)
        data_inefficient = pandas.read_excel(data_file, sheetname='%s_inefficient_sgRNA' % d, skiprows=2)

        data_efficient['threshold'] = 1.
        data_inefficient['threshold'] = 0.

        exp_data = pandas.concat((data_efficient, data_inefficient))
        exp_data['rank_KBM7'] = exp_data.groupby('Gene Symbol')['log2 fold change, KBM7'].transform(rank_transform)
        exp_data['rank_HL60'] = exp_data.groupby('Gene Symbol')['log2 fold change, HL60'].transform(rank_transform)

        if aggregated is None:
            aggregated = exp_data
        else:
            aggregated = pandas.concat((aggregated, exp_data))


    # go from 40mer to 30mer
    if subsetting == 'ours':
        aggregated["sequence(target+3'+5')"] = aggregated["sequence(target+3'+5')"].apply(lambda x: x[6:-4])
    else:
        aggregated["sequence(target+3'+5')"] = aggregated["sequence(target+3'+5')"].apply(lambda x: x[10:])

    # make sure EVEYTHING is uppercase
    aggregated["sequence(target+3'+5')"] = aggregated["sequence(target+3'+5')"].apply(lambda x: x.upper())

    # rename columns
    aggregated.rename(columns={"sequence(target+3'+5')": '30mer', 'Gene Symbol': 'Target gene', 'strand':'Strand'}, inplace=True)

    aggregated['Strand'].loc[aggregated['Strand']=='+'] = 'sense'
    aggregated['Strand'].loc[aggregated['Strand']=='-'] = 'antisense'

    aggregated['average rank'] = aggregated[['rank_HL60', 'rank_KBM7']].mean(axis=1)
    df = aggregated
    df = df.rename(columns={'30mer': 'Sequence', 'Target gene': 'Target'})
    df['drug'] = 'nodrug'
    df['test'] = 1
    df = df.set_index(['Sequence', 'Target', 'drug'])
    df['30mer'] = df.index.get_level_values(0)
    df['Target gene'] = df.index.get_level_values(1)
    df['Organism'] = 'unknown'
    df['score_drug_gene_rank'] = df['average rank']
    df['score_drug_gene_threshold'] = df['threshold']
    df['Nucleotide cut position'] = df['start of target']
    df['Percent Peptide'] = 0
    df['Amino Acid Cut position'] = 0
    target_genes = np.unique(df['Target gene'].values)

    return df[['Nucleotide cut position', 'Percent Peptide', 'Amino Acid Cut position']], target_genes, df[['30mer', 'Strand']], df[['score_drug_gene_rank', 'score_drug_gene_threshold', 'test', 'Target gene']]

def read_V2_data(data_file, learn_options=None, verbose=True):
    if data_file is None:
        data_file = cur_dir + "/data/V2_data.xlsx"

    # to compare
    # import predict as pr; a1, g1, t1, X1, Y1 = pr.data_setup()
    # a1.index.names

    data = pandas.read_excel(data_file, sheetname="ResultsFiltered", skiprows=list(range(0, 6+1)), index_col=[0, 4])
    # grab data relevant to each of three drugs, which exludes some genes
    # note gene MED12 has two drugs, all others have at most one
    Xdf = pandas.DataFrame()

    # This comes from the "Pairs" tab in their excel sheet,
    # note HPRT/HPRT1 are same thing, and also PLX_2uM/PLcX_2uM
    known_pairs = {'AZD_200nM':  ['CCDC101', 'MED12', 'TADA2B', 'TADA1'],
                   '6TG_2ug/mL': ['HPRT1'],
                   'PLX_2uM':    ['CUL3', 'NF1', 'NF2', 'MED12']}

    drugs_to_genes = {'AZD_200nM':  ['CCDC101', 'MED12', 'TADA2B', 'TADA1'],
                      '6TG_2ug/mL': ['HPRT1'],
                      'PLX_2uM':    ['CUL3', 'NF1', 'NF2', 'MED12']}

    if learn_options is not None:
        assert not (learn_options['extra pairs'] and learn_options['all pairs']), "extra pairs and all pairs options (in learn_options) can't be active simultaneously."

        if learn_options['extra pairs']:
            drugs_to_genes['AZD_200nM'].extend(['CUL3', 'NF1', 'NF2'])
        elif learn_options['all pairs']:
            drugs_to_genes['AZD_200nM'].extend(['HPRT1', 'CUL3', 'NF1', 'NF2'])
            drugs_to_genes['PLX_2uM'].extend(['HPRT1', 'CCDC101', 'TADA2B', 'TADA1'])
            drugs_to_genes['6TG_2ug/mL'].extend(['CCDC101', 'MED12', 'TADA2B', 'TADA1', 'CUL3', 'NF1', 'NF2'])

    count = 0
    for drug in list(drugs_to_genes.keys()):
        genes = drugs_to_genes[drug]
        for g in genes:
            Xtmp = data.copy().xs(g, level='Target gene', drop_level=False)
            Xtmp['drug'] = drug
            Xtmp['score'] = Xtmp[drug].copy()  # grab the drug results that are relevant for this gene

            if g in known_pairs[drug]:
                Xtmp['test'] = 1.
            else:
                Xtmp['test'] = 0.

            count = count + Xtmp.shape[0]
            Xdf = pandas.concat([Xdf, Xtmp], axis=0)
            if verbose:
                print("Loaded %d samples for gene %s \ttotal number of samples: %d" % (Xtmp.shape[0], g, count))

    # create new index that includes the drug
    Xdf = Xdf.set_index('drug', append=True)

    Y = pandas.DataFrame(Xdf.pop("score"))
    Y.columns.names = ["score"]

    test_gene = pandas.DataFrame(Xdf.pop('test'))
    target = pandas.DataFrame(Xdf.index.get_level_values('Target gene').values, index=Y.index, columns=["Target gene"])
    Y = pandas.concat((Y, target, test_gene), axis=1)
    target_genes = Y['Target gene'].unique()
    gene_position = Xdf[["Percent Peptide", "Amino Acid Cut position"]].copy()

    # convert to ranks for each (gene, drug combo)
    # flip = True
    y_rank = pandas.DataFrame()
    y_threshold = pandas.DataFrame()
    y_quant = pandas.DataFrame()
    for drug in list(drugs_to_genes.keys()):
        gene_list = drugs_to_genes[drug]
        for gene in gene_list:
            ytmp = pandas.DataFrame(Y.xs((gene, drug), level=["Target gene", "drug"], drop_level=False)['score'])
            y_ranktmp, y_rank_raw, y_thresholdtmp, y_quanttmp = util.get_ranks(ytmp, thresh=0.8, prefix="score_drug_gene", flip=False)
            # np.unique(y_rank.values-y_rank_raw.values)
            y_rank = pandas.concat((y_rank, y_ranktmp), axis=0)
            y_threshold = pandas.concat((y_threshold, y_thresholdtmp), axis=0)
            y_quant = pandas.concat((y_quant, y_quanttmp), axis=0)

    yall = pandas.concat((y_rank, y_threshold, y_quant), axis=1)
    Y = pandas.merge(Y, yall, how='inner', left_index=True, right_index=True)

    # convert also by drug only, irrespective of gene
    y_rank = pandas.DataFrame()
    y_threshold = pandas.DataFrame()
    y_quant = pandas.DataFrame()
    for drug in list(drugs_to_genes.keys()):
        ytmp = pandas.DataFrame(Y.xs(drug, level="drug", drop_level=False)['score'])
        y_ranktmp, y_rank_raw, y_thresholdtmp, y_quanttmp = util.get_ranks(ytmp, thresh=0.8, prefix="score_drug", flip=False)
        # np.unique(y_rank.values-y_rank_raw.values)
        y_rank = pandas.concat((y_rank, y_ranktmp), axis=0)
        y_threshold = pandas.concat((y_threshold, y_thresholdtmp), axis=0)
        y_quant = pandas.concat((y_quant, y_quanttmp), axis=0)

    yall = pandas.concat((y_rank, y_threshold, y_quant), axis=1)
    Y = pandas.merge(Y, yall, how='inner', left_index=True, right_index=True)

    PLOT = False
    if PLOT:
        # to better understand, try plotting something like:
        labels = ["score", "score_drug_gene_rank", "score_drug_rank", "score_drug_gene_threshold", "score_drug_threshold"]

        for label in labels:
            plt.figure()
            plt.plot(Xdf['sgRNA Score'].values, Y[label].values, '.')
            r, pearp = sp.stats.pearsonr(Xdf['sgRNA Score'].values.flatten(), Y[label].values.flatten())
            plt.title(label + ' VS pred. score, $r$=%0.2f (p=%0.2e)' % (r, pearp))
            plt.xlabel("sgRNA prediction score")
            plt.ylabel(label)

    gene_position = util.impute_gene_position(gene_position)

    if learn_options is not None and learn_options["weighted"] == "variance":
        print("computing weights from replicate variance...")
        # compute the variance across replicates so can use it as a weight
        data = pandas.read_excel(data_file, sheetname="Normalized", skiprows=list(range(0, 6+1)), index_col=[0, 4])
        data.index.names = ["Sequence", "Target gene"]

        experiments = {}
        experiments['AZD_200nM'] = ['Deep 25', 'Deep 27', 'Deep 29 ', 'Deep 31']
        experiments['6TG_2ug/mL'] = ['Deep 33', 'Deep 35', 'Deep 37', 'Deep 39']
        experiments['PLX_2uM'] = ['Deep 49', 'Deep 51', 'Deep 53', 'Deep 55']

        variance = None
        for drug in list(drugs_to_genes.keys()):
            data_tmp = data.iloc[data.index.get_level_values('Target gene').isin(drugs_to_genes[drug])][experiments[drug]]
            data_tmp["drug"] = drug
            data_tmp = data_tmp.set_index('drug', append=True)
            data_tmp["variance"] = np.var(data_tmp.values, axis=1)
            if variance is None:
                variance = data_tmp["variance"].copy()
            else:
                variance = pandas.concat((variance, data_tmp["variance"]), axis=0)

        orig_index = Y.index.copy()
        Y = pandas.merge(Y, pandas.DataFrame(variance), how="inner", left_index=True, right_index=True)
        Y = Y.ix[orig_index]
        print("done.")

    # Make sure to keep this check last in this function
    assert Xdf.index.equals(Y.index), "The index of Xdf is different from the index of Y (this can cause inconsistencies/random performance later on)"

    return Xdf, drugs_to_genes, target_genes, Y, gene_position


def merge_all(data_file=None, data_file2=None, data_file3=None, learn_options=None):
    Xdf, Y, gene_position, target_genes = mergeV1_V2(data_file, data_file2, learn_options)
    gene_position_xu, target_genes_xu, Xdf_xu, Y_xu = read_xu_et_al(data_file3, learn_options)
    Xdf = pandas.concat((Xdf, Xdf_xu))
    Y = pandas.concat((Y, Y_xu))
    gene_position = pandas.concat((gene_position, gene_position_xu))
    target_genes = np.concatenate((target_genes, target_genes_xu))

    return Xdf, Y, gene_position, target_genes

def mergeV1_V2(data_file, data_file2, learn_options):
    '''
    ground_truth_label, etc. are taken to correspond to the V2 data, and then the V1 is appropriately matched
    based on semantics
    '''
    assert not learn_options['include_strand'], "don't currently have 'Strand' column in V1 data"

    annotations, gene_position1, target_genes1, Xdf1, Y1 = read_V1_data(data_file, learn_options)
    Xdf2, drugs_to_genes, target_genes2, Y2, gene_position2 = read_V2_data(data_file2)

    Y1.rename(columns={'average rank': learn_options["rank-transformed target name"]}, inplace=True)
    Y1.rename(columns={'average threshold': learn_options["binary target name"]}, inplace=True)

    # rename columns, and add a dummy "drug" to V1 so can join the data sets
    Y1["drug"] = ["nodrug" for x in range(Y1.shape[0])]
    Y1 = Y1.set_index('drug', append=True)
    Y1.index.names = ['Sequence', 'Target gene', 'drug']

    Y_cols_to_keep = np.unique(['Target gene', 'test', 'score_drug_gene_rank', 'score_drug_gene_threshold'])

    Y1 = Y1[Y_cols_to_keep]
    Y2 = Y2[Y_cols_to_keep]

    Xdf1["drug"] = ["nodrug" for x in range(Xdf1.shape[0])]
    Xdf1 = Xdf1.set_index('drug', append=True)

    X_cols_to_keep = ['30mer', 'Strand']
    Xdf1 = Xdf1[X_cols_to_keep]
    Xdf2 = Xdf2[X_cols_to_keep]

    gene_position1["drug"] = ["nodrug" for x in range(gene_position1.shape[0])]
    gene_position1 = gene_position1.set_index('drug', append=True)
    gene_position1.index.names = ['Sequence', 'Target gene', 'drug']
    cols_to_keep = ['Percent Peptide', 'Amino Acid Cut position']
    gene_position1 = gene_position1[cols_to_keep]
    gene_position2 = gene_position2[cols_to_keep]

    Y = pandas.concat((Y1, Y2), axis=0)
    Xdf = pandas.concat((Xdf1, Xdf2), axis=0)
    gene_position = pandas.concat((gene_position1, gene_position2))

    # target_genes = target_genes1 + target_genes2
    target_genes = np.concatenate((target_genes1, target_genes2))

    save_to_file = False

    if save_to_file:
        Y.index.names = ['Sequence', 'Target', 'drug']
        assert np.all(Xdf.index.values==Y.index.values), "rows don't match up"

        onedupind = np.where(Y.index.duplicated())[0][0]
        alldupind = np.where(Y.index.get_level_values(0).values==Y.index[onedupind][0])[0]

        #arbitrarily set one of these to have "nodrug2" as the third level index
        #so that they are not repeated, and the joints therefore do not augment the data set
        assert len(alldupind)==2, "expected only duplicates"
        newindex = Y.index.tolist()
        newindex[onedupind] = (newindex[onedupind][0], newindex[onedupind][1], "nodrug2")
        Y.index = pandas.MultiIndex.from_tuples(newindex, names = Y.index.names)
        Xdf.index = pandas.MultiIndex.from_tuples(newindex, names = Y.index.names)

        # there seems to be a duplicate index, and thus this increases the data set size, so doing it the hacky way...
        XandY = pandas.merge(Xdf, Y, how="inner", left_index=True, right_index=True)
        gene_position_tmp = gene_position.copy()
        gene_position_tmp.index.names = ['Sequence', 'Target', 'drug']
        gene_position_tmp.index = pandas.MultiIndex.from_tuples(newindex, names = Y.index.names)
        XandY = pandas.merge(XandY, gene_position_tmp, how="inner", left_index=True, right_index=True)

        # truncate to 30mers
        XandY["30mer"] = XandY["30mer"].apply(lambda x: x[0:30])
        XandY.to_csv(r'D:\Source\CRISPR\data\tmp\V3.csv')

    return Xdf, Y, gene_position, target_genes


def get_V1_genes(data_file=None):
    annotations, gene_position, target_genes, Xdf, Y = read_V1_data(data_file, learn_options=None)
    return target_genes


def get_V2_genes(data_file=None):
    Xdf, drugs_to_genes, target_genes, Y, gene_position = read_V2_data(data_file, verbose=False)
    return target_genes


def get_V3_genes(data_fileV1=None, data_fileV2=None):
    target_genes = np.concatenate((get_V1_genes(data_fileV1), get_V2_genes(data_fileV2)))
    return target_genes

def get_xu_genes(data_file=None):
    return read_xu_et_al(data_file)[1]

def get_mouse_genes(data_file=None):
    annotations, gene_position, target_genes, Xdf, Y = read_V1_data(data_file, learn_options=None)
    return Xdf[Xdf['Organism'] == 'mouse']['Target gene'].unique()


def get_human_genes(data_file=None):
    annotations, gene_position, target_genes, Xdf, Y = read_V1_data(data_file, learn_options=None)
    mouse_genes = Xdf[Xdf['Organism'] == 'mouse']['Target gene'].unique()
    all_genes = get_V3_genes(None, None)  # TODO this needs to support specifying file names (!= 'None')
    return np.setdiff1d(all_genes, mouse_genes)



import numpy as np
import matplotlib.pyplot as plt


def gp_on_fold(feature_sets, train, test, y, y_all, inputs, dim, dimsum, learn_options):
    import GPy
    from models.gpy_ssk import WeightedDegree

    sequences = np.array([str(x) for x in y_all.index.get_level_values(0).tolist()])

    kern = WeightedDegree(1, sequences, d=learn_options['kernel degree'], active_dims=[0])
    X = np.arange(len(train))[:, None]

    current_dim = 1

    if 'gc_count' in feature_sets:
        kern += GPy.kern.RBF(1, active_dims=[current_dim], name='GC_rbf')
        X = np.concatenate((X, feature_sets['gc_count'].values), axis=1)
        current_dim += 1
        assert X.shape[1] == current_dim

    if 'drug' in feature_sets:
        Q = feature_sets['drug'].values.shape[1]
        kern += GPy.kern.Linear(Q, active_dims=list(range(current_dim, current_dim+Q)), name='drug_lin')
        X = np.concatenate((X, feature_sets['drug'].values), axis=1)
        current_dim += Q
        assert X.shape[1] == current_dim

    if 'gene effect' in feature_sets:
        Q = feature_sets['gene effect'].values.shape[1]
        kern += GPy.kern.Linear(Q, active_dims=list(range(current_dim, current_dim+Q)), name='gene_lin')
        X = np.concatenate((X, feature_sets['gene effect'].values), axis=1)
        current_dim += Q
        assert X.shape[1] == current_dim

    if "Percent Peptide" in feature_sets:
        Q = feature_sets['Percent Peptide'].values.shape[1]
        kern += GPy.kern.RBF(Q, active_dims=list(range(current_dim, current_dim+Q)), name='percent_pept')
        X = np.concatenate((X, feature_sets['Percent Peptide'].values), axis=1)
        current_dim += Q
        assert X.shape[1] == current_dim

    if "Nucleotide cut position" in feature_sets:
        Q = feature_sets['Nucleotide cut position'].values.shape[1]
        kern += GPy.kern.RBF(Q, active_dims=list(range(current_dim, current_dim+Q)), name='nucleo_cut')
        X = np.concatenate((X, feature_sets['Nucleotide cut position'].values), axis=1)
        current_dim += Q
        assert X.shape[1] == current_dim

    if "Strand effect" in feature_sets:
        Q = feature_sets['Strand effect'].values.shape[1]
        kern += GPy.kern.Linear(Q, active_dims=list(range(current_dim, current_dim+Q)), name='strand')
        X = np.concatenate((X, feature_sets['Strand effect'].values), axis=1)
        current_dim += Q
        assert X.shape[1] == current_dim

    if "NGGX" in feature_sets:
        Q = feature_sets['NGGX'].values.shape[1]
        kern += GPy.kern.Linear(Q, active_dims=list(range(current_dim, current_dim+Q)), name='NGGX')
        X = np.concatenate((X, feature_sets['NGGX'].values), axis=1)
        current_dim += Q
        assert X.shape[1] == current_dim

    if "TM" in feature_sets:
        Q = feature_sets['TM'].values.shape[1]
        kern += GPy.kern.RBF(Q, ARD=True, active_dims=list(range(current_dim, current_dim+Q)), name='TM')
        X = np.concatenate((X, feature_sets['TM'].values), axis=1)
        current_dim += Q
        assert X.shape[1] == current_dim

    if "gene features" in feature_sets:
        Q = feature_sets['gene features'].values.shape[1]
        kern += GPy.kern.Linear(Q, ARD=True, active_dims=list(range(current_dim, current_dim+Q)), name='genefeat')
        X = np.concatenate((X, feature_sets['gene features'].values), axis=1)
        current_dim += Q
        assert X.shape[1] == current_dim

    kern += GPy.kern.Bias(X.shape[1])

    if learn_options['warpedGP']:
        m = GPy.models.WarpedGP(X[train], y[train], kernel=kern)
    else:
        m = GPy.models.GPRegression(X[train], y[train], kernel=kern)


    if False: # debug plots
        m.optimize_restarts(3, messages=1)
        y_pred, y_uncert = m.predict(X[test])

        plt.figure('residuals')
        plt.title('residuals')
        plt.hist(y_pred.flatten()-y[test].flatten(), bins=50)

        plt.figure('predictions')
        plt.title('predictions')
        plt.errorbar(y_pred.flatten(), y[test].flatten(), yerr=np.sqrt(y_uncert.flatten()), fmt='o')
        plt.xlabel('prediction')
        plt.ylabel('truth')

        plt.figure('kernel')
        plt.title('kernel')
        plt.imshow(m.kern.K(X,X))
        print(m)
        print("%.3f variance explained" % (m.Gaussian_noise.variance/y[train].var()))
        import ipdb; ipdb.set_trace()
        plt.close('all')
    else:
        m.optimize_restarts(3)
        # m.optimize(messages=1)
        y_pred, y_uncert = m.predict(X[test])

    # TODO add offset such that low scores are around 0 (not -4 or so)

    return y_pred, m[:]



"""
Setup file for making the package pip installable
"""
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sgRNAble",
    version="1.0.10",
    author="Avery Noonan, Siddarth Raghuvanshi, Ahmed Abdelmoneim",
    author_email="sidr97@gmail.com",
    description="CRISPR-Cas9 single guide RNA generation tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Siddarth-Raghuvanshi/CRISPR-Guide-RNA",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    python_requires='==3.7.*',
    entry_points={
        "console_scripts": ['sgrnable = optimal_guide_finder.guide_finder:main']
    },
    install_requires=[
        "biopython",
        "matplotlib",
        "numba",
        "numpy==1.16.6",
        "pandas",
        "psutil",
        "scikit-learn==0.20.4",
        "scipy",
        "tqdm",
    ],
)



"""
Entry point to the program

- Accept and parse user commands
- Generate Potential Guides
- Run through biophysical model and report results
"""
import os
import logging
import argparse
import time
import psutil
import numpy as np
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from optimal_guide_finder import guide_generator, guide_strength_calculator, memory_limit

def init_parser():
    """
    Initializes a parser to get user arguments using argparse

    Returns:
        ArgumentParser -- parser ready to accept arguments
    """
    # Parser to get the files listed in the arguments
    parser = argparse.ArgumentParser(description="""This program helps you to find all possible guide RNAs that will
                                       target the gene. Then using the model created by Salis Lab,
                                       you can see the off target effects for the each possible guide.""",
                                     formatter_class=argparse.RawTextHelpFormatter)

    # Parsers to add arguements.
    parser.add_argument("-t", "--target_sequence", required=True,
                        help="The Gene Sequence of Interest (Fasta format)")
    parser.add_argument("-g", "--genome_sequence", required=True, nargs='+',
                        help="""The Genome of the organism, if targeting a plasmid, make sure to \n
                              include it as well (Fasta format)""")

    parser.add_argument("-a", "--azimuth_cutoff", type=int, required=False,
                        default=10, help="""How many guides should pass from azimuth screening,
                              the guides are passed based on descending azimuth prediction score""")
    parser.add_argument("-c", "--copy_number", required=False, default=1, nargs='+',
                        help="""Number of copies of target gene present""")
    parser.add_argument("-p", "--purpose", required=False, default="d",
                        help="""d: default, will look at all possible guides,
                            i: CRISPR interference on gene, will only look at negative strand""")

    parser.add_argument("-o", "--output_dir", required=False, default="output",
                        help="Path to output directory")

    parser.add_argument("-th", "--num_threads", required=False, default=None, type=int,
                        help="""Number of threads to use when running the program""")
    parser.add_argument("-m", "--max_memory", required=False, default=None, type=float,
                        help="""Maximum memory used by the tool in GiB,
                                defaults to using whatever memory is available in the system""")

    parser.add_argument("-v", "--verbose", action='store_true', help="""Enable verbose console logging""")

    return parser

def get_sequence(target_sequence, genome_sequences, copy_numbers):
    """
    Loads target and genome sequences from files, if multiple files are passed for genome sequences,
    they'll be combined into one file

    Args:
        target_sequence (string): path to fasta file containing target sequence
        genome_sequences (list(string)): list of paths to fasta files containing genome sequences
        copy_numbers (list(int)): number of copies of target sequence present for each target sequence

    Returns:
        [(dict, string)]: target sequence dictionary, genome sequence as string
    """
    # Reads the file using biopython and creates an object called target
    target_dict = SeqIO.to_dict(SeqIO.parse(target_sequence, "fasta"))

    for name in target_dict:
        target_dict[name] = target_dict[name].seq.upper()

    # Reads the Genome files using biopython and combines them into one genome object
    genome = SeqRecord(Seq(""))
    for i, genome_sequence in enumerate(genome_sequences):
        genome_parts = SeqIO.parse(genome_sequence, "fasta")
        for part in genome_parts:
            genome.seq = genome.seq + part.seq * int(copy_numbers[i])

    return target_dict, genome.seq.upper()

def initialize_logger(output_file, logging_level=logging.INFO):
    """
    Initialize logging to be used throughout the program

    Args:
        output_file (str): path to logging output
        logging_level (logging.LEVEL, optional): verbosity of console logger. Defaults to logging.INFO.
    """
    logging.basicConfig(level=logging.INFO,
                        filename=output_file + '/run.log',
                        filemode='w')
    console_logger = logging.StreamHandler()
    console_logger.setLevel(logging_level)
    logging.getLogger().addHandler(console_logger)

def main():
    """
    Main workflow
    """
    start_time = time.time()

    # init parser
    parser = init_parser()
    args = parser.parse_args()

    # create output directory if not already present
    if not os.path.isdir(args.output_dir):
        os.makedirs(args.output_dir)
    assert(os.path.isdir(args.output_dir))

    # set logger verbosity level
    if args.verbose:
        initialize_logger(args.output_dir)
    else:
        initialize_logger(args.output_dir, logging.ERROR)

    logger = logging.getLogger(__name__)

    # set memory limit if passed in
    if args.max_memory is not None:
        memory_limit.set_limit(args.max_memory)

    # create the path for genome file
    logger.info("Creating genome file...")
    genome_location = args.output_dir + '/Run_Genome'

    # Get the sequences in a Seq format from user fasta or genebank files
    logger.info("Generating target dictionary..")
    copies = [args.copy_number]*len(args.genome_sequence) if isinstance(args.copy_number, int) else args.copy_number
    target_dict, genome = get_sequence(args.target_sequence, args.genome_sequence, copies)

    ref_record = SeqRecord(genome, id="refgenome", name="reference", description="a reference background")
    ref_record = ref_record + ref_record.reverse_complement()
    SeqIO.write(ref_record, genome_location, "fasta")

    # Select the guides based on the purpose and the azimuth model
    logger.info("Selecting initial guides..")
    guide_list = guide_generator.select_guides(target_dict, args.purpose, args.azimuth_cutoff)

    # Build and run the model
    logger.info("Initializing binding strength model..")
    model = guide_strength_calculator.initalize_model(genome_location)

    logger.info("Processing guides off-target binding...")
    results_df = guide_strength_calculator.process_guides(model, guide_list, num_threads=args.num_threads)

    #generate and append Rank array
    logger.info("Generating result file..")
    results_df.sort_values(by=['Gene/ORF Name', 'Entropy Score'], inplace=True)
    results_df.drop_duplicates(inplace=True)
    rank_array = []
    for gene in results_df['Gene/ORF Name'].unique():
        num_guides = len(results_df[results_df['Gene/ORF Name'] == gene]['Guide Sequence'])
        rank_array.extend(list(np.arange(1, num_guides+1)))
    results_df['Rank in Target Gene'] = rank_array

    results_df.to_csv(args.output_dir + '/output.csv', index=False)

    process = psutil.Process(os.getpid())

    total_time = time.time() - start_time
    minutes = int(total_time//60)
    seconds = int(total_time%60)
    max_memory = process.memory_info().rss // (1024 * 1024)
    logger.info("Finished sgRNAble run, total time elapsed (min/sec): %d:%d, max memory usage (MB): %d", minutes, seconds, max_memory)

if __name__ == "__main__":
    main()



"""
Module for generating list of potential guides
"""
import numpy as np
from tqdm import tqdm
from Bio.Seq import Seq
from optimal_guide_finder.Azimuth_Model import model_comparison

PAM = "GG"
GUIDE_RNA_LENGTH = 20
AZIMUTH_DISTANCE = 24
TOTAL_AZIMUTH_DISTANCE = 30


def select_guides(target_dict, purpose, azimuth_cutoff):
    """
    Args:
        target_dict (dict): target sequences
        purpose (string): one of (d, i) for default or interference screen
        azimuth_cutoff (int): number of guides to consider from azimuth

    Returns:
        [dict]: list of guides indexed by target in target_dict
    """
    if purpose == "d":
        return _select_guides_default(target_dict, azimuth_cutoff)

    if purpose == "i":
        return _select_guides_interference(target_dict, azimuth_cutoff)

    raise ValueError(f"Unknown purpose {purpose}, must be one of (d, i)")

def _select_guides_default(target_dict, azimuth_cutoff):
    """
    Run guide selection for an default screen by considering all guides

    Args:
        target_dict (dict): target sequences
        azimuth_cutoff (int): number of guides to consider from azimuth

    Returns:
        [dict]: list of guides indexed by target in target_dict
    """
    guide_list = {}

    for gene in tqdm(target_dict, "Generating guides"):
        # Get guides on the positive strand
        guide_list[gene] = []
        locations = _find_pams(target_dict[gene])
        strand_array = ["Positive"] * len(locations)

        # Get guides on the negative strand
        neg_locations = _find_pams(Seq.reverse_complement(target_dict[gene]))
        sequence_length = len(target_dict[gene])
        neg_locations = [sequence_length - x for x in neg_locations]
        negative_array = ["Negative"] * len(neg_locations)

        # Combine the information into a single array for each gene
        # Magic numbers increase the sequence by 6 bps after NGG and 4 to raise total to 30
        guides = [str(target_dict[gene][loc-4:loc+GUIDE_RNA_LENGTH+6]) for loc in locations]

        # Magic numbers perform same as above, but in the opposite direction due to negative strand
        guides.extend([str(Seq.reverse_complement(
            target_dict[gene][loc-6-GUIDE_RNA_LENGTH:loc+4])) for loc in neg_locations])
        locations.extend(neg_locations)
        strand_array.extend(negative_array)

        #In case the gene is too short to have no Pam sites
        if not guides:
            guide_list.pop(gene)
            continue

        # Run the model through the Azimuth Model
        predictions = model_comparison.predict(np.array(guides))
        guides = [x for y, x in sorted(zip(predictions, guides), reverse=True)][:azimuth_cutoff]
        locations = [x for y, x in sorted(zip(predictions, locations), reverse=True)][:azimuth_cutoff]
        strand_array = [x for y, x in sorted(zip(predictions, strand_array), reverse=True)][:azimuth_cutoff]
        guides = [guide[4:24] for guide in guides]
        guide_list[gene].extend([guides, locations, strand_array])

    return guide_list

def _select_guides_interference(target_dict, azimuth_cutoff):
    """
    Run guide selection for an interference screen by only considering the negative strand

    Args:
        target_dict (dict): target sequences
        azimuth_cutoff (int): number of guides to consider from azimuth

    Returns:
        [dict]: list of guides indexed by target in target_dict
    """
    guide_list = {}

    for gene in tqdm(target_dict, "Generating guides"):
        # Get guides on the negative strand
        guide_list[gene] = []
        locations = _find_pams(Seq.reverse_complement(target_dict[gene]))
        sequence_length = len(target_dict[gene])
        locations = [sequence_length - x for x in locations]
        strand_array = ["Negative"] * len(locations)

        # Combine the information into a single array for each gene
        # Magic numbers increase the sequence by 6 bps after NGG and 4 to raise total to 30
        guides = ([str(Seq.reverse_complement(
            target_dict[gene][loc-6-GUIDE_RNA_LENGTH:loc+4])) for loc in locations])

        #In case the gene is too short to have no Pam sites
        if not guides:
            guide_list.pop(gene)
            continue

        # Run the model through the Azimuth Model
        predictions = model_comparison.predict(np.array(guides))
        guides = [x for y, x in sorted(zip(predictions, guides), reverse=True)][:azimuth_cutoff]
        locations = [x for y, x in sorted(zip(predictions, locations), reverse=True)][:azimuth_cutoff]
        strand_array = [x for y, x in sorted(zip(predictions, strand_array), reverse=True)][:azimuth_cutoff]
        guides = [guide[4:24] for guide in guides]
        guide_list[gene].extend([guides, locations, strand_array])

    return guide_list

def _find_pams(sequence):
    """
    Find all potential target in sequence containing PAM
    Args:
        sequence (string): target sequence

    Returns:
        [list(int)]: list of locations in sequence containing PAM sequence
    """
    locations = []
    sequence = str(sequence)
    position = 0

    while position < len(sequence):
        i = sequence[position:].find(PAM)
        if i < 0:
            break

        potential_guide_location = position + i - GUIDE_RNA_LENGTH

        # check if the guide is too close to the beginning of the gene.
        # Plus one is due to the N in the GG so the entire frame should be shifted down
        if potential_guide_location < (AZIMUTH_DISTANCE - GUIDE_RNA_LENGTH) + 1:
            pass

        # check if the guide is long enough for azimuth analysis
        elif potential_guide_location + TOTAL_AZIMUTH_DISTANCE < len(sequence):
            # the negative 1 accounts for the N in the GG
            locations.append(potential_guide_location - 1)

        position = position + i + 1

    return locations
